#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import numpy as _np

from coremltools.converters.mil.frontend.tensorflow.ops import (
    _transpose_NHWC_to_NCHW,
    _transpose_NCHW_to_NHWC,
)

# TF 2.x now imports and registers all TF 1.x op against the new registry
# (separated from TF 1.x registry). Overwrite might needed in case the op
# semantics are different between TF 1.x and TF 2.x.<
from coremltools.converters.mil.frontend.tensorflow.ops import *
from coremltools.converters.mil.frontend.tensorflow.dialect_ops import *


@register_tf_op(override=True)
def FusedBatchNormV3(context, node):
    # Get attributes
    data_format = node.attr.get("data_format", "NHWC")
    epsilon = node.attr.get("epsilon", None)

    # Get inputs
    x = context[node.inputs[0]]
    scale = context[node.inputs[1]]
    offset = context[node.inputs[2]]
    mean = context[node.inputs[3]]
    variance = context[node.inputs[4]]
    if data_format == "NHWC":
        # TF's FusedBatchNorm is only for 4D inputs
        x = _transpose_NHWC_to_NCHW(x)
        x = mb.batch_norm(
            x=x, mean=mean, variance=variance, gamma=scale, beta=offset, epsilon=epsilon
        )
        x = _transpose_NCHW_to_NHWC(x, node.name)
    else:
        x = mb.batch_norm(
            x=x,
            mean=mean,
            variance=variance,
            gamma=scale,
            beta=offset,
            epsilon=epsilon,
            name=node.name,
        )
    # Inference only batch norm does not have meaningful outputs for
    # batch_mean, batch_variance etc.
    context.add(node.name, x)


@register_tf_op(tf_alias=["If"], override=True)
def StatelessIf(context, node):
    pred = context[node.inputs[0]][0]
    then_graph = context.get_graph(node.attr.get("then_branch"))
    else_graph = context.get_graph(node.attr.get("else_branch"))

    def then_fn():
        context.stack_func_inputs(context[node.inputs[0]])
        then_output_var = convert_graph(context, then_graph)
        context.unstack_func_inputs()
        return then_output_var

    def else_fn():
        context.stack_func_inputs(context[node.inputs[0]])
        else_output_var = convert_graph(context, else_graph)
        context.unstack_func_inputs()
        return else_output_var

    x = mb.cond(pred=pred, _true_fn=then_fn, _false_fn=else_fn, name=node.name)

    # wraps x as tuple for get_tuple that always follow the cond node.
    x = (x,) if not isinstance(x, (tuple, list)) else x

    context.add(node.name, x)


@register_tf_op(tf_alias=["While"], override=True)
def StatelessWhile(context, node):
    # inputs are loop_counter, max_iterations, [loop_vars]
    loop_vars = context[node.inputs[0]][2:]

    cond_graph = context.get_graph(node.attr.get("cond"))
    body_graph = context.get_graph(node.attr.get("body"))

    def cond(*loop_vars):
        context.stack_func_inputs(loop_vars)
        cond_output_vars = convert_graph(context, cond_graph)
        context.unstack_func_inputs()
        return cond_output_vars

    def body(*loop_vars):
        context.stack_func_inputs(loop_vars)
        body_output_vars = convert_graph(context, body_graph)
        context.unstack_func_inputs()
        return body_output_vars

    x = mb.while_loop(_cond=cond, _body=body, loop_vars=loop_vars, name=node.name)

    # wraps x as tuple for get_tuple that always follow the while node.
    x = (x,) if not isinstance(x, (tuple, list)) else x

    context.add(node.name, x)


@register_tf_op
def TensorListFromTensor(context, node):
    value = context[node.inputs[0]]
    element_shape = context[node.inputs[1]]
    element_dtype = node.attr.get("element_dtype")
    dtype_str = types.builtin_to_string(element_dtype)

    length = mb.shape(x=value)
    length = mb.slice_by_index(x=length, begin=[0], end=[1], squeeze_mask=[True])

    if element_shape is not None and all(_np.atleast_1d(element_shape.val) != -1):
        ls = mb.make_list(init_length=length,
            elem_shape=tuple(element_shape.val.tolist()), dtype=dtype_str)
    else:
        ls = mb.tf_make_list(init_length=length, dtype=dtype_str)

    indices = mb.range_1d(end=length, start=0, step=1)
    ls = mb.list_scatter(ls=ls, indices=indices, value=value, name=node.name)
    context.add(node.name, ls)


@register_tf_op
def TensorListGather(context, node):
    ls = context[node.inputs[0]]
    indices = context[node.inputs[1]]
    tensor = mb.list_gather(ls=ls, indices=indices, name=node.name)
    context.add(node.name, tensor)


@register_tf_op
def TensorListGetItem(context, node):
    ls = context[node.inputs[0]]
    index = context[node.inputs[1]]
    new_ls = mb.list_read(ls=ls, index=index, name=node.name)
    context.add(node.name, new_ls)


@register_tf_op
def TensorListLength(context, node):
    ls = context[node.inputs[0]]
    length = mb.list_length(ls=ls, name=node.name)
    context.add(node.name, length)


@register_tf_op
def TensorListResize(context, node):
    # skip here as the list will be dynamically resized when
    # necessary in downstream list_write or list_scatter ops
    Identity(context, node)


@register_tf_op
def TensorListReserve(context, node):
    element_shape = context[node.inputs[0]]
    num_elements = context[node.inputs[1]]
    element_dtype = node.attr.get("element_dtype")
    dtype = types.builtin_to_string(element_dtype)

    if element_shape is not None and all(_np.atleast_1d(element_shape.val) != -1):
        ls = mb.make_list(
            init_length=num_elements,
            elem_shape=tuple(element_shape.val.tolist()),
            dtype=dtype,
            name=node.name,
        )
    else:
        ls = mb.tf_make_list(init_length=num_elements, dtype=dtype, name=node.name)
    context.add(node.name, ls)


@register_tf_op
def TensorListScatterIntoExistingList(context, node):
    ls = context[node.inputs[0]]
    value = context[node.inputs[1]]
    indices = context[node.inputs[2]]
    ls = mb.list_scatter(ls=ls, indices=indices, value=value, name=node.name)
    context.add(node.name, ls)


@register_tf_op
def TensorListSetItem(context, node):
    ls = context[node.inputs[0]]
    index = context[node.inputs[1]]
    value = context[node.inputs[2]]
    new_ls = mb.list_write(ls=ls, index=index, value=value, name=node.name)
    context.add(node.name, new_ls)


@register_tf_op
def TensorListStack(context, node):
    ls = context[node.inputs[0]]
    length = mb.list_length(ls=ls)
    indices = mb.range_1d(end=length, start=0, step=1)
    x = mb.list_gather(ls=ls, indices=indices, name=node.name)
    context.add(node.name, x)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.testing_reqs import *

from .testing_utils import run_compare_builder

backends = testing_reqs.backends


class TestAvgPool:
    @pytest.mark.parametrize(
        "use_cpu_only, backend, num_dims",
        itertools.product([True, False], backends, [1, 2, 3]),
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend, num_dims):
        kernel_sizes = [1, 2, 3]
        strides = [2, 1, 3]

        if num_dims == 1:
            x_val = np.array([[[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]]], dtype=np.float32)
            expected_output_types = [(1, 1, 4, types.fp32), (1, 1, 3, types.fp32)]
            expected_outputs = [
                np.array([[[1.0, 3.0, 5.0, 7.0]]], dtype=np.float32),
                np.array([[[1.5, 4.0, 6.5]]], dtype=np.float32),
            ]
        elif num_dims == 2:
            x_val = np.array(
                [
                    [
                        [[-10.80291205, -6.42076184], [-7.07910997, 9.1913279]],
                        [[-3.18181497, 0.9132147], [11.9785544, 7.92449539]],
                    ]
                ],
                dtype=np.float32,
            )
            expected_output_types = [(1, 2, 1, 1, types.fp32), (1, 2, 2, 1, types.fp32)]
            expected_outputs = [
                np.array([[[[-8.611837]], [[-1.1343001]]]], dtype=np.float32),
                np.array(
                    [[[[-3.7778642], [1.056109]], [[4.4086123], [9.951525]]]],
                    dtype=np.float32,
                ),
            ]
        else:  # num_dims == 3
            x_val = np.array(
                [
                    [
                        [
                            [[-1, -5, -1], [-3, -3, 8], [2, 6, 2]],
                            [[-4, 7, -4], [4, 6, 7], [4, 4, 8]],
                            [[5, -3, 5], [0, -5, 8], [1, 7, 2]],
                        ]
                    ],
                    [
                        [
                            [[7, -3, -5], [5, 4, 7], [-2, -4, -3]],
                            [[-4, 3, -1], [6, -4, 4], [3, 6, 2]],
                            [[-1, 4, -4], [-2, -1, -2], [3, 2, 8]],
                        ]
                    ],
                ],
                dtype=np.float32,
            )
            expected_output_types = [
                (2, 1, 2, 2, 1, types.fp32),
                (2, 1, 2, 3, 1, types.fp32),
            ]
            expected_outputs = [
                np.array(
                    [
                        [[[[-0.8333334], [2.0]], [[1.6666667], [2.1666667]]]],
                        [[[[2.5], [1.1666667]], [[-1.0], [1.3333334]]]],
                    ],
                    dtype=np.float32,
                ),
                np.array(
                    [
                        [
                            [
                                [[-0.8333334], [2.0], [3.3333335]],
                                [[1.6666667], [2.1666667], [3.3333335]],
                            ]
                        ],
                        [
                            [
                                [[2.5], [1.1666667], [-3.0]],
                                [[-1.0], [1.3333334], [4.3333335]],
                            ]
                        ],
                    ],
                    dtype=np.float32,
                ),
            ]

        input_values = {"x": x_val}
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}

        def build(x):
            return [
                mb.avg_pool(
                    x=x,
                    kernel_sizes=kernel_sizes[:num_dims],
                    strides=strides[:num_dims],
                    pad_type="valid",
                ),
                mb.avg_pool(
                    x=x,
                    kernel_sizes=kernel_sizes[-num_dims:],
                    strides=strides[-num_dims:],
                    pad_type="same",
                    exclude_padding_from_average=True,
                ),
            ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )


class TestMaxPool:
    @pytest.mark.parametrize(
        "use_cpu_only, backend, num_dims",
        itertools.product([True, False], backends, [1, 2, 3]),
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend, num_dims):
        kernel_sizes = [1, 2, 3]
        strides = [2, 1, 3]

        if num_dims == 1:
            x_val = np.array([[[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]]], dtype=np.float32)
            expected_output_types = [(1, 1, 4, types.fp32), (1, 1, 3, types.fp32)]
            expected_outputs = [
                np.array([[[1.0, 3.0, 5.0, 7.0]]], dtype=np.float32),
                np.array([[[2.0, 5.0, 7.0]]], dtype=np.float32),
            ]
        elif num_dims == 2:
            x_val = np.array(
                [
                    [
                        [[-10.80291205, -6.42076184], [-7.07910997, 9.1913279]],
                        [[-3.18181497, 0.9132147], [11.9785544, 7.92449539]],
                    ]
                ],
                dtype=np.float32,
            )
            expected_output_types = [(1, 2, 1, 1, types.fp32), (1, 2, 2, 1, types.fp32)]
            expected_outputs = [
                np.array([[[[-6.42076184]], [[0.9132147]]]], dtype=np.float32),
                np.array(
                    [[[[9.191328], [9.191328]], [[11.978555], [11.978555]]]],
                    dtype=np.float32,
                ),
            ]
        else:  # num_dims == 3
            x_val = np.array(
                [
                    [
                        [
                            [[-1, -5, -1], [-3, -3, 8], [2, 6, 2]],
                            [[-4, 7, -4], [4, 6, 7], [4, 4, 8]],
                            [[5, -3, 5], [0, -5, 8], [1, 7, 2]],
                        ]
                    ],
                    [
                        [
                            [[7, -3, -5], [5, 4, 7], [-2, -4, -3]],
                            [[-4, 3, -1], [6, -4, 4], [3, 6, 2]],
                            [[-1, 4, -4], [-2, -1, -2], [3, 2, 8]],
                        ]
                    ],
                ],
                dtype=np.float32,
            )
            expected_output_types = [
                (2, 1, 2, 2, 1, types.fp32),
                (2, 1, 2, 3, 1, types.fp32),
            ]
            expected_outputs = [
                np.array(
                    [
                        [[[[8.0], [8.0]], [[8.0], [8.0]]]],
                        [[[[7.0], [7.0]], [[4.0], [8.0]]]],
                    ],
                    dtype=np.float32,
                ),
                np.array(
                    [
                        [[[[8.0], [8.0], [6.0]], [[8.0], [8.0], [7.0]]]],
                        [[[[7.0], [7.0], [-2.0]], [[4.0], [8.0], [8.0]]]],
                    ],
                    dtype=np.float32,
                ),
            ]

        input_values = {"x": x_val}
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}

        def build(x):
            return [
                mb.max_pool(
                    x=x,
                    kernel_sizes=kernel_sizes[:num_dims],
                    strides=strides[:num_dims],
                    pad_type="valid",
                ),
                mb.max_pool(
                    x=x,
                    kernel_sizes=kernel_sizes[-num_dims:],
                    strides=strides[-num_dims:],
                    pad_type="same",
                ),
            ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )


class TestL2Pool:
    @pytest.mark.parametrize(
        "use_cpu_only, backend, num_dims",
        itertools.product([True, False], backends, [1, 2]),
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend, num_dims):
        kernel_sizes = [1, 2, 3]
        strides = [2, 1, 3]

        if num_dims == 1:
            x_val = np.array([[[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]]], dtype=np.float32)
            expected_output_types = [(1, 1, 4, types.fp32), (1, 1, 3, types.fp32)]
            expected_outputs = [
                np.array([[[1.0, 3.0, 5.0, 7.0]]], dtype=np.float32),
                np.array([[[2.236068, 7.071068, 9.219544]]], dtype=np.float32),
            ]
        elif num_dims == 2:
            x_val = np.array(
                [[[[-10.0, -6.0], [-7.0, 9.0]], [[-3.0, 0.0], [11.0, 7.0]]]],
                dtype=np.float32,
            )
            expected_output_types = [(1, 2, 1, 1, types.fp32), (1, 2, 2, 1, types.fp32)]
            expected_outputs = [
                np.array([[[[11.66190338]], [[3.0]]]], dtype=np.float32),
                np.array(
                    [[[[16.309507], [11.401754]], [[13.379088], [13.038404]]]],
                    dtype=np.float32,
                ),
            ]
        else:  # num_dims == 3
            pass  # Enum PoolingType3D has no value defined for name L2

        input_values = {"x": x_val}
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}

        def build(x):
            return [
                mb.l2_pool(
                    x=x,
                    kernel_sizes=kernel_sizes[:num_dims],
                    strides=strides[:num_dims],
                    pad_type="valid",
                ),
                mb.l2_pool(
                    x=x,
                    kernel_sizes=kernel_sizes[-num_dims:],
                    strides=strides[-num_dims:],
                    pad_type="same",
                ),
            ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
import logging


def dead_code_elimination_block(block):
    used_vars = set()
    ops_to_remove = list()

    # mark block's outputs to used
    used_vars.update(block.outputs)

    for op in reversed(block.operations):
        # if none of op's output is used, delete op
        if not set(op.outputs).intersection(used_vars):
            ops_to_remove.append(op)
            continue

        # mark all op's inputs to used
        for _, input_var in op.inputs.items():
            if isinstance(input_var, (tuple, list)):
                used_vars.update(list(input_var))
            else:
                used_vars.update([input_var])

        for b in op.blocks:
            used_in_block = dead_code_elimination_block(b)
            used_vars.update(used_in_block)

    for op in ops_to_remove:
        logging.info('Removing op "{}" (type: {})'.format(op.name, op.op_type))
        op.remove_from_block()

    return used_vars


@register_pass(namespace="common")
def dead_code_elimination(program):
    """
    Eliminate unused ops in program.

    Parameters
    ----------
    program: Program SSA Program before graph pass

    Returns
    -------
    program: Program SSA Program after graph pass

    Example
    -------

        Given:
        main(%x: (2, 4, fp32)) {
          block0() {
            %const_2: (4, 2, fp32)* = const(val=[...])
            %const_3: (4, fp32)* = const(val=[...])
            %tx_0: (bool)* = const(val=False)
            %ty_0: (bool)* = const(val=False)
            %matmul_0: (2, 2, fp32) = matmul(x=%x, y=%const_2, transpose_x=%tx_0, transpose_y=%ty_0)
            %linear_0: (2, 4, fp32) = linear(x=%x, weight=%const_2, bias=%const_3)
          } -> (%linear_0)
        }

        Result:
        main(%x: (2, 4, fp32)) {
          block0() {
            %const_2: (4, 2, fp32)* = const(val=[...])
            %const_3: (4, fp32)* = const(val=[...])
            %linear_0: (2, 4, fp32) = linear(x=%x, weight=%const_2, bias=%const_3)
          } -> (%linear_0)
        }

    Ops whose outputs are not contributed to final outputs will be deleted.
    In this example, %matmul_0 is an op that's not used in the computation,
    this op and its input ops (%tx_0 and %ty_0) are eliminated in this pass.
    """

    for name, f in program.functions.items():
        dead_code_elimination_block(f)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import scipy
from scipy import special
from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.testing_reqs import *

from .testing_utils import run_compare_builder

backends = testing_reqs.backends


class TestElementwiseUnary:
    # All ops in this test share the same backends
    @pytest.mark.parametrize(
        "use_cpu_only, backend, mode",
        itertools.product(
            [True, False],
            backends,
            [
                "abs",
                "acos",
                "asin",
                "atan",
                "atanh",
                "exp2",
                "clip",
                "cos",
                "cosh",
                "erf",
                "exp",
                "erf",
                "floor",
                "inverse",
                "log",
                "round",
                "rsqrt",
                "sign",
                "sin",
                "sinh",
                "sqrt",
                "square",
                "tan",
                "tanh",
                "threshold",
                "cast",
            ],
        ),
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend, mode):
        if mode == "abs":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)

            build = lambda x: mb.abs(x=x)
        elif mode == "acos":
            val = np.array([[-1, -0.5, 0], [0.4, 0.5, 0.8]], dtype=np.float32)
            expected_outputs = np.array(
                [
                    [3.14159265, 2.0943951, 1.57079633],
                    [1.15927948, 1.04719755, 0.64350111],
                ],
                dtype=np.float32,
            )

            build = lambda x: mb.acos(x=x)
        elif mode == "asin":
            val = np.array([[-1, -0.5, 0], [0.4, 0.5, 0.8]], dtype=np.float32)
            expected_outputs = np.array(
                [[-1.57079633, -0.52359878, 0.0], [0.41151685, 0.52359878, 0.92729522]],
                dtype=np.float32,
            )

            build = lambda x: mb.asin(x=x)
        elif mode == "atan":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [
                    [-0.78539816, 1.10714872, -1.24904577],
                    [1.32581766, -1.37340077, 1.40564765],
                ],
                dtype=np.float32,
            )
            build = lambda x: mb.atan(x=x)
        elif mode == "atanh":
            val = np.array([[-0.8, -0.5, 0], [0.4, 0.5, 0.8]], dtype=np.float32)
            expected_outputs = np.array(
                [[-1.09861229, -0.54930614, 0.0], [0.42364893, 0.54930614, 1.09861229]],
                dtype=np.float32,
            )

            build = lambda x: mb.atanh(x=x)
        elif mode == "cast":
            val = np.array([[-1.2, 2, -3.6], [4.5, -5, 6.7]], dtype=np.float32)
            expected_outputs = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.int32)
            build = lambda x: mb.cast(x=x, dtype="int32")
        elif mode == "ceil":
            val = np.array([[-1.2, 2, -3.4], [4.5, -5, 6.7]], dtype=np.float32)
            expected_outputs = np.array([[-1, 2, -3], [5, -5, 7]], dtype=np.float32)

            build = lambda x: mb.ceil(x=x)
        elif mode == "clip":
            val = np.array([[-1.2, 2, -3.4], [4.5, -5, 6.7]], dtype=np.float32)
            expected_outputs = np.array([[0, 2, 0], [4.5, 0, 5]], dtype=np.float32)

            build = lambda x: mb.clip(x=x, alpha=0.0, beta=5.0)
        elif mode == "cos":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [
                    [0.54030231, -0.41614684, -0.9899925],
                    [-0.65364362, 0.28366219, 0.96017029],
                ],
                dtype=np.float32,
            )

            build = lambda x: mb.cos(x=x)
        elif mode == "cosh":
            val = np.array([[-1, -2, -3], [1, 2, 3]], dtype=np.float32)
            expected_outputs = np.array(
                [
                    [1.54308063, 3.76219569, 10.067662],
                    [1.54308063, 3.76219569, 10.067662],
                ],
                dtype=np.float32,
            )

            build = lambda x: mb.cosh(x=x)
        elif mode == "erf":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [
                    [-0.8427007929497148, 0.9953222650189527, -0.9999779095030014],
                    [0.9999999845827421, -0.9999999999984626, 1.0],
                ],
                dtype=np.float32,
            )

            build = lambda x: mb.erf(x=x)
        elif mode == "exp":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [
                    [0.36787944, 7.3890561, 0.04978707],
                    [54.5981500, 0.0067379, 403.428793],
                ],
                dtype=np.float32,
            )

            build = lambda x: mb.exp(x=x)
        elif mode == "exp2":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [[0.5, 4.0, 0.125], [16, 0.03125, 64]], dtype=np.float32
            )

            build = lambda x: mb.exp2(x=x)
        elif mode == "floor":
            val = np.array([[-1.2, 2, -3.4], [4.5, -5, 6.7]], dtype=np.float32)
            expected_outputs = np.array([[-2, 2, -4], [4, -5, 6]], dtype=np.float32)

            build = lambda x: mb.floor(x=x)
        elif mode == "inverse":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [[-1.0, 0.5, -0.33333334], [0.25, -0.2, 0.16666667]], dtype=np.float32
            )
            build = lambda x: mb.inverse(x=x)
        elif mode == "log":
            val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [[0.0, 0.69314718, 1.09861229], [1.38629436, 1.60943791, 1.79175947]],
                dtype=np.float32,
            )

            build = lambda x: mb.log(x=x)
        elif mode == "round":
            val = np.array([[-1.2, 2, -3.4], [4.6, -5, 6.7]], dtype=np.float32)
            expected_outputs = np.array([[-1, 2, -3], [5, -5, 7]], dtype=np.float32)

            build = lambda x: mb.round(x=x)
        elif mode == "rsqrt":
            val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [[1.0, 0.70710678, 0.57735027], [0.5, 0.4472136, 0.40824829]],
                dtype=np.float32,
            )

            build = lambda x: mb.rsqrt(x=x)
        elif mode == "sign":
            val = np.array([[-1, 2, 0], [0, -5, 6]], dtype=np.float32)
            expected_outputs = np.array([[-1, 1, 0], [0, -1, 1]], dtype=np.float32)

            build = lambda x: mb.sign(x=x)
        elif mode == "sin":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [
                    [-0.84147098, 0.90929743, -0.14112001],
                    [-0.7568025, 0.95892427, -0.2794155],
                ],
                dtype=np.float32,
            )

            build = lambda x: mb.sin(x=x)
        elif mode == "sinh":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [[-1.1752, 3.62686, -10.017874], [27.289917, -74.20321, 201.71315]],
                dtype=np.float32,
            )

            build = lambda x: mb.sinh(x=x)
        elif mode == "sqrt":
            val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [[1.0, 1.41421356, 1.73205081], [2.0, 2.23606798, 2.44948974]],
                dtype=np.float32,
            )

            build = lambda x: mb.sqrt(x=x)
        elif mode == "square":
            val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [[1.0, 4.0, 9.0], [16.0, 25.0, 36.]],
                dtype=np.float32,
            )

            build = lambda x: mb.square(x=x)
        elif mode == "tan":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [[-1.5574, -2.185, 0.1425], [1.15782, 3.3805, -0.291]], dtype=np.float32
            )

            build = lambda x: mb.tan(x=x)
        elif mode == "tanh":
            val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [
                    [-0.7615942, 0.9640276, -0.9950548],
                    [0.9993293, -0.9999092, 0.9999877],
                ],
                dtype=np.float32,
            )

            build = lambda x: mb.tanh(x=x)
        elif mode == "threshold":
            val = np.array([[-1.2, 2, -3.4], [4.5, -5, 6.7]], dtype=np.float32)
            expected_outputs = np.array(
                [[1.0, 2, 1.0], [4.5, 1.0, 6.7]], dtype=np.float32
            )

            build = lambda x: mb.threshold(x=x, alpha=1.0)

        input_placeholders = {"x": mb.placeholder(shape=val.shape)}
        input_values = {"x": val}
        expected_output_types = (
            (2, 3, types.int32) if mode == "cast" else (2, 3, types.fp32)
        )

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_abs_eval(self):
        val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.abs(x=val)
        expected_outputs = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_acos_eval(self):
        val = np.array([[-1, -0.5, 0], [0.4, 0.5, 0.8]], dtype=np.float32)
        v = mb.acos(x=val)
        expected_outputs = np.array(
            [[3.14159265, 2.0943951, 1.57079633], [1.15927948, 1.04719755, 0.64350111]],
            dtype=np.float32,
        )
        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_asin_eval(self):
        val = np.array([[-1, -0.5, 0], [0.4, 0.5, 0.8]], dtype=np.float32)
        v = mb.asin(x=val)
        expected_outputs = np.array(
            [[-1.57079633, -0.52359878, 0.0], [0.41151685, 0.52359878, 0.92729522]],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_atan_eval(self):
        val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.atan(x=val)
        expected_outputs = np.array(
            [
                [-0.78539816, 1.10714872, -1.24904577],
                [1.32581766, -1.37340077, 1.40564765],
            ],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_atanh_eval(self):
        val = np.array([[-0.8, -0.5, 0], [0.4, 0.5, 0.8]], dtype=np.float32)
        v = mb.atanh(x=val)
        expected_outputs = np.array(
            [[-1.09861229, -0.54930614, 0.0], [0.42364893, 0.54930614, 1.09861229]],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_cast_eval(self):
        val = np.array([[-1.2, 2, -3.4], [4.5, -5, 6.7]], dtype=np.float32)
        expected_outputs = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.int32)

        v = mb.cast(x=val, dtype="int32")

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_ceil_eval(self):
        val = np.array([[-1.2, 2, -3.4], [4.5, -5, 6.7]], dtype=np.float32)
        v = mb.ceil(x=val)
        expected_outputs = np.array([[-1, 2, -3], [5, -5, 7]], dtype=np.float32)

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_clip_eval(self):
        val = np.array([[-1.2, 2, -3.4], [4.5, -5, 6.7]], dtype=np.float32)
        v = mb.clip(x=val, alpha=0.0, beta=5.0)
        expected_outputs = np.array([[0, 2, 0], [4.5, 0, 5]], dtype=np.float32)

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_cos_eval(self):
        val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.cos(x=val)
        expected_outputs = np.array(
            [
                [0.54030231, -0.41614684, -0.9899925],
                [-0.65364362, 0.28366219, 0.96017029],
            ],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_cosh_eval(self):
        val = np.array([[-1, -2, -3], [1, 2, 3]], dtype=np.float32)
        v = mb.cosh(x=val)
        expected_outputs = np.array(
            [[1.54308063, 3.76219569, 10.067662], [1.54308063, 3.76219569, 10.067662]],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_erf_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.erf(x=x_val)
        assert is_close(scipy.special.erf(x_val), v.val)

    @ssa_fn
    def test_builder_exp_eval(self):
        val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.exp(x=val)
        expected_outputs = np.array(
            [[0.36787944, 7.3890561, 0.04978707], [54.5981500, 0.0067379, 403.428793]],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_exp2_eval(self):
        val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.exp2(x=val)
        expected_outputs = np.array(
            [[0.5, 4.0, 0.125], [16, 0.03125, 64]], dtype=np.float32
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_floor_eval(self):
        val = np.array([[-1.2, 2, -3.4], [4.5, -5, 6.7]], dtype=np.float32)
        v = mb.floor(x=val)
        expected_outputs = np.array([[-2, 2, -4], [4, -5, 6]], dtype=np.float32)

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_inverse_eval(self):
        val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.inverse(x=val)
        expected_outputs = np.array(
            [[-1.0, 0.5, -0.33333334], [0.25, -0.2, 0.16666667]], dtype=np.float32
        )
        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_log_eval(self):
        val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        v = mb.log(x=val)
        expected_outputs = np.array(
            [[0.0, 0.69314718, 1.09861229], [1.38629436, 1.60943791, 1.79175947]],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_round_eval(self):
        val = np.array([[-1.2, 2, -3.4], [4.6, -5, 6.7]], dtype=np.float32)
        v = mb.round(x=val)
        expected_outputs = np.array([[-1, 2, -3], [5, -5, 7]], dtype=np.float32)

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_rsqrt_eval(self):
        val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        v = mb.rsqrt(x=val)
        expected_outputs = np.array(
            [[1.0, 0.70710678, 0.57735027], [0.5, 0.4472136, 0.40824829]],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_sign_eval(self):
        val = np.array([[-1, 2, 0], [0, -5, 6]], dtype=np.float32)
        v = mb.sign(x=val)
        expected_outputs = np.array([[-1, 1, 0], [0, -1, 1]], dtype=np.float32)

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_sin_eval(self):
        val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.sin(x=val)
        expected_outputs = np.array(
            [
                [-0.84147098, 0.90929743, -0.14112001],
                [-0.7568025, 0.95892427, -0.2794155],
            ],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_sinh_eval(self):
        val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.sinh(x=val)
        expected_outputs = np.array(
            [[-1.1752, 3.62686, -10.017874], [27.289917, -74.20321, 201.71315]],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_sqrt_eval(self):
        val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        v = mb.sqrt(x=val)
        expected_outputs = np.array(
            [[1.0, 1.41421356, 1.73205081], [2.0, 2.23606798, 2.44948974]],
            dtype=np.float32,
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_tan_eval(self):
        val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.tan(x=val)
        expected_outputs = np.array(
            [[-1.5574, -2.185, 0.1425], [1.15782, 3.3805, -0.291]], dtype=np.float32
        )

        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_tanh_eval(self):
        x_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.tanh(x=x_val)
        assert is_close(np.tanh(x_val), v.val)

    @ssa_fn
    def test_builder_threshold_eval(self):
        val = np.array([[-1.2, 2, -3.4], [4.5, -5, 6.7]], dtype=np.float32)
        v = mb.threshold(x=val, alpha=1.0)
        expected_outputs = np.array([[1.0, 2, 1.0], [4.5, 1.0, 6.7]], dtype=np.float32)

        assert is_close(expected_outputs, v.val)

    @pytest.mark.parametrize(
        "use_cpu_only, backend, epsilon",
        itertools.product(
            [True, False],
            backends,
            [1e-3, 1e-1, 1.0],
        ),
    )
    def test_builder_to_backend_stress_inverse(
        self, use_cpu_only, backend, epsilon
    ):
        x = np.array([[1, -2, 3], [4, -5, 6]], dtype=np.float32)
        numpy_pred = 1 / (x + epsilon)

        input_placeholder_dict = {"x": mb.placeholder(shape=x.shape)}
        input_value_dict = {"x": x}

        def build(x):
            return mb.inverse(x=x, epsilon=epsilon)

        expected_output_type = x.shape + (types.fp32,)
        run_compare_builder(
            build,
            input_placeholder_dict,
            input_value_dict,
            expected_output_type,
            numpy_pred,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, epsilon",
        itertools.product(
            [True, False],
            backends,
            [1e-3, 1e-1, 1.0],
        ),
    )
    def test_builder_to_backend_stress_rsqrt(
        self, use_cpu_only, backend, epsilon
    ):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        numpy_pred = 1.0 / np.sqrt(x + epsilon)

        input_placeholder_dict = {"x": mb.placeholder(shape=x.shape)}
        input_value_dict = {"x": x}

        def build(x):
            return mb.rsqrt(x=x, epsilon=epsilon)

        expected_output_type = x.shape + (types.fp32,)
        run_compare_builder(
            build,
            input_placeholder_dict,
            input_value_dict,
            expected_output_type,
            numpy_pred,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, epsilon",
        itertools.product(
            [True, False],
            backends,
            [1e-3, 1e-1, 1.0],
        ),
    )
    def test_builder_to_backend_stress_log(
            self, use_cpu_only, backend, epsilon
    ):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        numpy_pred = np.log(x + epsilon)

        input_placeholder_dict = {"x": mb.placeholder(shape=x.shape)}
        input_value_dict = {"x": x}

        def build(x):
            return mb.log(x=x, epsilon=epsilon)

        expected_output_type = x.shape + (types.fp32,)
        run_compare_builder(
            build,
            input_placeholder_dict,
            input_value_dict,
            expected_output_type,
            numpy_pred,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.testing_utils import (
    assert_op_count_match,
    assert_model_is_valid,
    get_op_types_in_program,
    apply_pass_and_basic_check,
)

import pytest
import numpy as np
import itertools

np.random.seed(1984)


class TestElementwiseOptimizationPasses:
    """
    Input graph:
                                    Const
                                      |
                                      V
    input -----> convolution -----> add/sub  ----> relu ---> out

    Output graph:
    input -----> convolution -----> relu ----> out
    """

    @pytest.mark.parametrize(
        "conv_dim, \
                             flip_add_input_order, \
                             add_batch_dim_to_const, \
                             use_sub_instead, \
                             prebuilt_bias, \
                             scalar_elementwise, \
                             use_conv_transpose",
        itertools.product(
            [
                2,
                3,
            ],  # 1D conv conversion broken even without the pass: rdar://problem/62960720
            [True, False],  # flip_add_input_order
            [True, False],  # add_batch_dim_to_const
            [True, False],  # use_sub_instead
            [True, False],  # prebuilt_bias
            [True, False],  # scalar_elementwise
            [True, False],  # use_conv_transpose
        ),
    )
    def test_fuse_bias_conv(
        self,
        conv_dim,
        flip_add_input_order,
        add_batch_dim_to_const,
        use_sub_instead,
        prebuilt_bias,
        scalar_elementwise,
        use_conv_transpose,
    ):

        if flip_add_input_order and use_sub_instead:
            return

        if use_conv_transpose and conv_dim != 2:
            return

        input_shape = None
        W = None
        Cout = 8
        Cin = 3
        D = 10
        const = (
            np.random.rand(Cout) if add_batch_dim_to_const else np.random.rand(1, Cout)
        )
        const = np.expand_dims(const, axis=-1)

        if conv_dim == 1:
            input_shape = (1, Cin, D)
            W = np.random.rand(Cout, Cin, 1)
        elif conv_dim == 2:
            input_shape = (1, Cin, D, D)
            W = np.random.rand(Cout, Cin, 1, 1)
            const = np.expand_dims(const, axis=-1)
        elif conv_dim == 3:
            input_shape = (1, Cin, D, D, D)
            W = np.random.rand(Cout, Cin, 1, 1, 1)
            const = np.expand_dims(const, axis=-1)
            const = np.expand_dims(const, axis=-1)

        output_shape = list(input_shape)
        output_shape[1] = Cout

        if scalar_elementwise:
            const = np.random.uniform(0)

        @mb.program(input_specs=[mb.TensorSpec(shape=input_shape)])
        def prog(x):
            kwargs = {
                "x": x,
                "weight": W,
                "pad_type": "valid",
                "dilations": [1] * conv_dim,
                "strides": [1] * conv_dim,
            }
            if prebuilt_bias:
                kwargs["bias"] = np.random.rand(Cout)

            x = mb.conv_transpose(**kwargs) if use_conv_transpose else mb.conv(**kwargs)

            if use_sub_instead:
                x = mb.sub(x=x, y=const)
            else:
                x = mb.add(
                    x=const if flip_add_input_order else x,
                    y=x if flip_add_input_order else const,
                )
            x = mb.relu(x=x)
            return x

        element_op = "sub" if use_sub_instead else "add"
        conv_op = "conv" if not use_conv_transpose else "conv_transpose"

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::fuse_bias_conv"
        )
        assert get_op_types_in_program(prev_prog) == [conv_op, element_op, "relu"]
        assert get_op_types_in_program(prog) == [conv_op, "relu"]

        old_bias = prev_block.find_ops(op_type=conv_op)[0].inputs.get("bias", None)
        old_bias_val = 0 if old_bias is None else old_bias.val
        assert old_bias_val is not None
        assert block.find_ops(op_type=conv_op)[0].inputs["bias"] is not None
        new_bias_val = block.find_ops(op_type=conv_op)[0].inputs["bias"].val
        assert new_bias_val is not None
        if use_sub_instead:
            np.testing.assert_almost_equal(
                old_bias_val - np.squeeze(const), new_bias_val
            )
        else:
            np.testing.assert_almost_equal(
                old_bias_val + np.squeeze(const), new_bias_val
            )

        assert_model_is_valid(
            prog,
            {"x": input_shape},
            expected_output_shapes={block.outputs[0].name: tuple(output_shape)},
        )

    """
    Input graph:
                                 Const     Const
                                   |         |
                                   V         V
    input -----> transpose -----> mul ----> add ---> out

    Output graph:
    input -----> transpose -----> batchnorm ----> out
    """

    @pytest.mark.parametrize(
        "flip_mul_input_order, flip_add_input_order, rank_3_const_input",
        itertools.product([False, True], [False, True], [False, True]),
    )
    def test_mul_add_fusion_to_batchnorm(
        self, flip_mul_input_order, flip_add_input_order, rank_3_const_input
    ):

        C = 3
        gamma = np.random.rand(1, C, 1, 1)
        beta = np.random.rand(1, C, 1, 1)
        if rank_3_const_input:
            gamma = np.squeeze(gamma, axis=0)
            beta = np.squeeze(beta, axis=0)

        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 10, 10, C))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            if flip_mul_input_order:
                x = mb.mul(x=gamma, y=x)
            else:
                x = mb.mul(x=x, y=gamma)
            if flip_add_input_order:
                x = mb.add(x=beta, y=x)
            else:
                x = mb.add(x=x, y=beta)
            return x

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::fuse_elementwise_to_batchnorm"
        )
        assert get_op_types_in_program(prev_prog) == ["transpose", "mul", "add"]
        assert get_op_types_in_program(prog) == ["transpose", "batch_norm"]
        assert_model_is_valid(
            prog,
            {"x": (1, 10, 10, C)},
            expected_output_shapes={block.outputs[0].name: (1, C, 10, 10)},
        )



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _


class Type:
    """
     - Type.name : A string with the name of the object
     - Type.tparam : For classes with template parameters, (list, dict), this
         contains a list of Type objects of the template parameters
     - Type.python_class : The original python class implementing this type.
                         Two Type objects compare equal
                         only on name and tparam and not python_class
    """

    __slots__ = ["name", "tparam", "python_class"]

    def __init__(self, name, tparam=None, python_class=None):
        if tparam is None:
            tparam = []
        assert isinstance(name, str)
        assert isinstance(tparam, list)
        self.name = name
        self.tparam = tparam
        self.python_class = python_class

    def __hash__(self):
        return hash((self.name, tuple(self.tparam)))

    def __eq__(self, other):
        return self.name == other.name and self.tparam == other.tparam

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        ret = self.name
        if len(self.tparam) > 0:
            ret += "[" + ",".join(repr(x) for x in self.tparam) + "]"
        return ret

    def __str__(self):
        return self.__repr__()

    def sexp(self):
        if len(self.tparam) == 0:
            return self.name
        else:
            ret = [self.name]
            ret.append([a.sexp() if hasattr(a, "sexp") else a for a in self.tparam])
            return ret


class FunctionType:
    """
    - FunctionType.inputs : A list of Type objects defining the types of the input
    - FunctionType.output: A Type object defining the type of the output
    - FunctionType.python_function : The original python function implementing
                                     this type. Two FunctionType objects compare
                                     equal only on inputs and output and not
                                     python_function
    """

    __slots__ = ["inputs", "output", "python_function"]

    def __init__(self, inputs, output, python_function=None):
        assert isinstance(inputs, list)
        assert isinstance(output, (FunctionType, Type))
        self.inputs = inputs
        self.output = output
        self.python_function = python_function

    def __hash__(self):
        return hash((tuple(self.inputs), self.output))

    def __eq__(self, other):
        return self.inputs == other.inputs and self.output == other.output

    def __repr__(self):
        return "(" + ",".join(repr(x) for x in self.inputs) + ")->" + repr(self.output)

    def __str__(self):
        return self.__repr__()

    def return_sexp(self):
        return self.output.sexp()

    def inputs_sexp(self):
        return [i.sexp() for i in self.inputs]



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import copy
from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.mil.passes.pass_registry import PASS_REGISTRY
from coremltools.converters.mil.testing_utils import assert_model_is_valid
from coremltools.converters.mil.testing_utils import assert_same_output_names


def test_commingle_loop_vars():
    def body(a, b):
        # b is a loop invariant
        return mb.add(x=a, y=b), b

    def cond(a, b):
        a_mean = mb.reduce_mean(x=a, axes=[0, 1])
        b_mean = mb.reduce_mean(x=b, axes=[0, 1])
        return mb.less(x=a_mean, y=b_mean)

    @mb.program(
        input_specs=[mb.TensorSpec(shape=(1, 2)), mb.TensorSpec(shape=(1, 2)),]
    )
    def prog(a, b):
        return mb.while_loop(_cond=cond, _body=body, loop_vars=(a, b))

    while_op = prog.find_ops(op_type="while_loop", exactly_one=True)[0]
    assert while_op.blocks[0].inputs[0].name == "a_x1"
    assert while_op.blocks[0].inputs[1].name == "b_x1"

    prev_prog = copy.deepcopy(prog)
    PASS_REGISTRY["nn_backend::commingle_loop_vars"](prog)
    assert_same_output_names(prev_prog, prog)

    while_op = prog.find_ops(op_type="while_loop", exactly_one=True)[0]
    assert while_op.blocks[0].inputs[0].name == while_op.outputs[0].name
    assert while_op.blocks[0].inputs[1].name == while_op.outputs[1].name

    prog.validate()

    # The program is not ssa and thus cannot be converted


def test_handle_return_return_inputs_as_outputs():
    @mb.program(
        input_specs=[mb.TensorSpec(shape=(1, 2)), mb.TensorSpec(shape=(1, 2)),]
    )
    def prog(a, b):
        return mb.mul(x=a, y=2), b

    prev_main_output_names = [o.name for o in prog["main"].outputs]
    assert prog["main"].outputs[1].op is None  # output comes from input

    prev_prog = copy.deepcopy(prog)
    PASS_REGISTRY["nn_backend::handle_return_inputs_as_outputs"](prog)
    assert_same_output_names(prev_prog, prog)

    assert prog["main"].outputs[1].op is not None  # output comes from an op
    assert prog["main"].outputs[1].op.op_type == "identity"

    assert_model_is_valid(prog, {"a": (1, 2), "b": (1, 2)})


def test_handle_unused_inputs():
    @mb.program(
        input_specs=[mb.TensorSpec(shape=(1, 2)),]
    )
    def prog(unused_input):
        return mb.const(val=[3, 2])

    prev_prog = copy.deepcopy(prog)
    PASS_REGISTRY["nn_backend::handle_unused_inputs"](prog)
    assert_same_output_names(prev_prog, prog)

    id_op = prog.find_ops(op_type="identity", exactly_one=True)[0]
    # Assert that input var is consumed by an identity op.
    assert id_op in prog["main"].inputs["unused_input"].child_ops

    assert_model_is_valid(prog, {"unused_input": (1, 2)})



from __future__ import print_function as _
from ...proto import NeuralNetwork_pb2 as _NeuralNetwork_pb2


def _get_weight_param_summary(wp):
    """Get a summary of _NeuralNetwork_pb2.WeightParams
    Args:
    wp : _NeuralNetwork_pb2.WeightParams - the _NeuralNetwork_pb2.WeightParams message to display
    Returns:
    a str summary for wp
    """
    summary_str = ""
    if wp.HasField("quantization"):
        nbits = wp.quantization.numberOfBits
        quant_type = (
            "linearly"
            if wp.quantization.HasField("linearQuantization")
            else "lookup-table"
        )
        summary_str += "{}-bit {} quantized".format(nbits, quant_type)

    if len(wp.floatValue) > 0:
        summary_str += "({} floatValues)".format(len(wp.floatValue))
    if len(wp.float16Value) > 0:
        summary_str += "({} bytes float16Values)".format(len(wp.float16Value))
    if len(wp.rawValue) > 0:
        summary_str += "({} bytes rawValues)".format(len(wp.rawValue))

    return summary_str


def _get_lstm_weight_param_summary(lstm_wp):
    weight_name_list = [
        "W_i",
        "W_f",
        "W_z",
        "W_o",
        "H_i",
        "H_f",
        "H_z",
        "H_o",
        "b_i",
        "b_f",
        "b_z",
        "b_o",
        "p_i",
        "p_f",
        "p_o",
    ]
    wp_summary_list = [
        _get_weight_param_summary(lstm_wp.inputGateWeightMatrix),
        _get_weight_param_summary(lstm_wp.forgetGateWeightMatrix),
        _get_weight_param_summary(lstm_wp.blockInputWeightMatrix),
        _get_weight_param_summary(lstm_wp.outputGateWeightMatrix),
        _get_weight_param_summary(lstm_wp.inputGateRecursionMatrix),
        _get_weight_param_summary(lstm_wp.forgetGateRecursionMatrix),
        _get_weight_param_summary(lstm_wp.blockInputRecursionMatrix),
        _get_weight_param_summary(lstm_wp.outputGateRecursionMatrix),
        _get_weight_param_summary(lstm_wp.inputGateBiasVector),
        _get_weight_param_summary(lstm_wp.forgetGateBiasVector),
        _get_weight_param_summary(lstm_wp.blockInputBiasVector),
        _get_weight_param_summary(lstm_wp.outputGateBiasVector),
        _get_weight_param_summary(lstm_wp.inputGatePeepholeVector),
        _get_weight_param_summary(lstm_wp.forgetGatePeepholeVector),
        _get_weight_param_summary(lstm_wp.outputGatePeepholeVector),
    ]
    lstm_wp_summary_list = []
    for idx, summary in enumerate(wp_summary_list):
        if len(summary) > 0:
            lstm_wp_summary_list.append(weight_name_list[idx] + ", " + summary)

    return ("\n" + " " * 8).join(lstm_wp_summary_list)


def _get_feature_description_summary(feature):
    if feature.type.HasField("multiArrayType"):
        shape = list(feature.type.multiArrayType.shape)
        int_shape = [int(x) for x in shape]
        return str(int_shape)
    else:
        return ("({})".format(str(feature.type))).replace("\n", "")


def _summarize_network_layer_info(layer):
    """
    Args:
    layer - an MLModel NeuralNetwork Layer protobuf message
    Returns:
    layer_type : str - type of layer
    layer_name : str - name of the layer
    layer_inputs : list[str] - a list of strings representing input blobs of the layer
    layer_outputs : list[str] - a list of strings representing output blobs of the layer
    layer_field_content : list[(str, str)] - a list of two-tuple of (parameter_name, content)
    """
    layer_type_str = layer.WhichOneof("layer")

    layer_name = layer.name
    layer_inputs = list(layer.input)
    layer_outputs = list(layer.output)

    typed_layer = getattr(layer, layer_type_str)
    layer_field_names = [l.name for l in typed_layer.DESCRIPTOR.fields]
    layer_field_content = []

    for name in layer_field_names:
        field = getattr(typed_layer, name)
        summary_str = ""
        if type(field) == _NeuralNetwork_pb2.LSTMWeightParams:
            summary_str = _get_lstm_weight_param_summary(field)
        elif type(field) == _NeuralNetwork_pb2.WeightParams:
            summary_str = _get_weight_param_summary(field)
        else:
            field_str = str(field)
            if len(field_str) > 0:
                summary_str = field_str.replace("\n", " ")
        if len(summary_str) > 0:
            layer_field_content.append([name, summary_str])

    return layer_type_str, layer_name, layer_inputs, layer_outputs, layer_field_content


def _summarize_neural_network_spec(mlmodel_spec):
    """ Summarize network into the following structure.
    Args:
    mlmodel_spec : mlmodel spec
    Returns:
    inputs : list[(str, str)] - a list of two tuple (name, descriptor) for each input blob.
    outputs : list[(str, str)] - a list of two tuple (name, descriptor) for each output blob
    layers : list[(str, list[str], list[str], list[(str, str)])] - a list of layers represented by
        layer name, input blobs, output blobs, a list of (parameter name, content)
    """
    inputs = [
        (blob.name, _get_feature_description_summary(blob))
        for blob in mlmodel_spec.description.input
    ]
    outputs = [
        (blob.name, _get_feature_description_summary(blob))
        for blob in mlmodel_spec.description.output
    ]
    nn = None

    if mlmodel_spec.HasField("neuralNetwork"):
        nn = mlmodel_spec.neuralNetwork
    elif mlmodel_spec.HasField("neuralNetworkClassifier"):
        nn = mlmodel_spec.neuralNetworkClassifier
    elif mlmodel_spec.HasField("neuralNetworkRegressor"):
        nn = mlmodel_spec.neuralNetworkRegressor

    layers = (
        [_summarize_network_layer_info(layer) for layer in nn.layers]
        if nn != None
        else None
    )
    return (inputs, outputs, layers)


def _prRed(skk, end=None):
    print("\033[91m {}\033[00m".format(skk), end=end)


def _prLightPurple(skk, end=None):
    print("\033[94m {}\033[00m".format(skk), end=end)


def _prPurple(skk, end=None):
    print("\033[95m {}\033[00m".format(skk), end=end)


def _prGreen(skk, end=None):
    print("\033[92m {}\033[00m".format(skk), end=end)


def _print_layer_type_and_arguments(
    layer_type_str, layer_inputs, indentation, to_indent=True, shape=None, value=None
):
    if to_indent:
        _prRed(indentation * "\t" + "{}".format(layer_type_str), end="")
    else:
        _prRed("{}".format(layer_type_str), end="")

    if shape is None:
        _prLightPurple("({})".format(", ".join(layer_inputs)))
    elif value is not None:
        _prLightPurple("(shape = ", end="")
        print("{}, ".format(str(shape)), end="")
        _prLightPurple("value = ", end="")
        values = ",".join(["{0: 0.1f}".format(v) for v in value]).lstrip()
        print("[{}]".format(values), end="")
        _prLightPurple(")")
    else:
        _prLightPurple("(shape = ", end="")
        print("{}".format(str(shape)), end="")
        _prLightPurple(")")


def _find_size(arr):
    s = 1
    for a in arr:
        s *= a
    return s


def _summarize_neural_network_spec_code_style(
    nn_spec, indentation=0, input_names=None, output_names=None
):
    """
    print nn_spec as if writing code
    """
    indentation_size = 1

    if input_names:
        print("def model({}):".format(", ".join(input_names)))
        indentation += indentation_size

    for i, layer in enumerate(nn_spec.layers):
        layer_type_str = layer.WhichOneof("layer")
        layer_inputs = list(layer.input)
        layer_outputs = list(layer.output)

        if layer_type_str == "loop":
            if len(layer.loop.conditionNetwork.layers) > 0:
                _prPurple(indentation * "\t" + "Condition Network: ")
                _summarize_neural_network_spec_code_style(
                    layer.loop.conditionNetwork, indentation=indentation
                )
            if layer.loop.conditionVar:
                layer_inputs.append(layer.loop.conditionVar)
            _print_layer_type_and_arguments(layer_type_str, layer_inputs, indentation)
            indentation += indentation_size
            _summarize_neural_network_spec_code_style(
                layer.loop.bodyNetwork, indentation=indentation
            )
            if len(layer.loop.conditionNetwork.layers) > 0:
                _prPurple(indentation * "\t" + "Condition Network: ")
                _summarize_neural_network_spec_code_style(
                    layer.loop.conditionNetwork, indentation=indentation
                )
            indentation -= indentation_size
            continue

        if layer_type_str == "branch":
            _print_layer_type_and_arguments(layer_type_str, layer_inputs, indentation)
            _prRed(indentation * "\t" + "IfBranch:")
            indentation += indentation_size
            _summarize_neural_network_spec_code_style(
                layer.branch.ifBranch, indentation=indentation
            )
            indentation -= indentation_size
            if len(layer.branch.elseBranch.layers) > 0:
                _prRed(indentation * "\t" + "ElseBranch:")
                indentation += indentation_size
                _summarize_neural_network_spec_code_style(
                    layer.branch.elseBranch, indentation=indentation
                )
                indentation -= indentation_size
            continue

        if layer_type_str == "loopBreak" or layer_type_str == "loopContinue":
            _prRed(indentation * "\t" + layer_type_str)
            continue

        shape = None
        value = None
        if layer_type_str == "loadConstant":
            shape = layer.loadConstant.shape
            shape = list(shape)
            int_shape = [int(x) for x in shape]
            shape = tuple([1, 1] + int_shape)
            size = _find_size(shape)
            if size < 4 and len(layer.loadConstant.data.floatValue) > 0:
                value = map(float, list(layer.loadConstant.data.floatValue))

        if layer_type_str == "loadConstantND":
            shape = layer.loadConstantND.shape
            shape = tuple(map(int, list(shape)))
            size = _find_size(shape)
            if size < 4 and len(layer.loadConstantND.data.floatValue) > 0:
                value = map(float, list(layer.loadConstantND.data.floatValue))

        print(indentation * "\t", end="")
        print("{} =".format(", ".join(layer_outputs)), end="")
        _print_layer_type_and_arguments(
            layer_type_str,
            layer_inputs,
            indentation,
            to_indent=False,
            shape=shape,
            value=value,
        )

    if output_names:
        _prRed("\n" + indentation * "\t" + "return ", end="")
        print("{}".format(", ".join(output_names)))



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import itertools

import numpy as np
import pytest
import torch
import torch.nn as nn

import coremltools
from .testing_utils import *

# Custom layer imports

from coremltools.converters.mil.mil.ops.defs._op_reqs import *
from coremltools.converters.mil.frontend.torch.torch_op_registry import (
    register_torch_op,
)
from coremltools.converters.mil.frontend.torch.torch_op_registry import (
    _TORCH_OPS_REGISTRY as _TORCH_OPS_REG,
)
from coremltools.converters.mil.frontend.torch.ops import _get_inputs
from coremltools.converters.mil.mil import Builder as mb

# Log Converter supported Cosine Similarity conversion function
default_cosine_similarity = _TORCH_OPS_REG.get("cosine_similarity", None)


@register_torch_op(override=True)
def cosine_similarity(context, node):
    inputs = _get_inputs(context, node, expected=4)
    dim = inputs[-2].val
    eps = inputs[-1].val
    xy = mb.mul(x=inputs[0], y=inputs[1])
    sum_xy = mb.reduce_sum(x=xy, axes=[dim])

    xx = mb.mul(x=inputs[0], y=inputs[0])
    sum_xx = mb.reduce_sum(x=xx, axes=[dim])
    yy = mb.mul(x=inputs[1], y=inputs[1])
    sum_yy = mb.reduce_sum(x=yy, axes=[dim])

    mul_sum_xy = mb.mul(x=sum_xx, y=sum_yy)
    div_12 = mb.maximum(x=mul_sum_xy, y=eps * eps)
    div_sqrt = mb.sqrt(x=div_12)

    cs = mb.real_div(x=sum_xy, y=div_sqrt, name=node.name)
    context.add(cs)


# Log custom Cosine Similarity conversion function
custom_cosine_similarity = _TORCH_OPS_REG["cosine_similarity"]


def _set_torch_reg_op(op_type, op_func):
    _TORCH_OPS_REG[op_type] = op_func


class TestCompositeOp(TorchBaseTest):

    @pytest.mark.xfail(reason="rdar://65230439 fails with stochastic numeric mismatch", run=False)
    @pytest.mark.parametrize("input_shape", [(100, 180), (56, 123)])
    def test_composite_op(self, input_shape):
        _set_torch_reg_op("cosine_similarity", custom_cosine_similarity)
        model = nn.CosineSimilarity(dim=1, eps=1e-6)
        self.run_compare_torch([input_shape, input_shape], model)
        _set_torch_reg_op("cosine_similarity", default_cosine_similarity)


class TestCustomOp:
    # Define SSA Custom Op for Sparse MatMul
    # This will map to `custom_op` in SSA with binding information
    # to bind input spec to the custom implementation
    @register_op(doc_str="Sparse MatMul Layer", is_custom_op=True)
    class custom_torch_sparse_matmul(Operation):
        # Defining input spec for current op
        input_spec = InputSpec(
            x=TensorInputType(),
            y=TensorInputType(),
            transpose_x=BoolInputType(const=True, optional=True),
            transpose_y=BoolInputType(const=True, optional=True),
            x_is_sparse=BoolInputType(const=True, optional=True),
            y_is_sparse=BoolInputType(const=True, optional=True),
        )

        def default_inputs(self):
            return DefaultInputs(
                transpose_x=False,
                transpose_y=False,
                x_is_sparse=False,
                y_is_sparse=False,
                )

        # Specifying binding for custom op for specifying inputs,
        # parameters required for creating custom op to be synced with Swift API
        bindings = {
            "class_name": "SparseMatMul",
            "input_order": ["x", "y"],
            "parameters": ["transpose_x", "transpose_y", "x_is_sparse", "y_is_sparse"],
            "description": "Custom Sparse MatMul Layer",
        }

        def __init__(self, **kwargs):
            super(TestCustomOp.custom_torch_sparse_matmul, self).__init__(**kwargs)

        def type_inference(self):
            x_type = self.x.dtype
            x_shape = self.x.shape
            y_shape = self.y.shape
            # For illustration purpose, assumming getting valid shape
            # Ideally, should consider transpose_?, ?_is_sparse parameters into consideration
            # for computing output shape
            ret_shape = [x_shape[0], y_shape[1]]
            return types.tensor(x_type, [x_shape[0], y_shape[1]])

    @register_torch_op()
    def _sparse_mm(context, node):
        inputs = _get_inputs(context, node, expected=2)
        x = mb.custom_torch_sparse_matmul(
            x=inputs[0], y=inputs[1], x_is_sparse=True, y_is_sparse=True, name=node.name
        )
        context.add(x)

    def test_custom_sparse_mm_op(self, input_shape=(4, 4)):
        class TestLayer(nn.Module):
            def __init__(self):
                super(TestLayer, self).__init__()

            def forward(self, x, y):
                x = torch.sparse.mm(x, y)
                return x

        model = TestLayer()
        input_data_x = torch.ones(input_shape)
        input_data_y = torch.ones(input_shape)
        input_data = [input_data_x, input_data_y]
        model.eval()
        torch_model = torch.jit.trace(model, (input_data_x, input_data_y))
        mlmodel = convert_to_mlmodel(torch_model, input_data)

        layers = mlmodel.get_spec().neuralNetwork.layers
        assert layers[-1].custom is not None, "Expecting a custom layer"
        assert (
            "SparseMatMul" == layers[-1].custom.className
        ), "Custom Layer class name mis-match"
        assert (
            False == layers[-1].custom.parameters["transpose_x"].boolValue
        ), "Incorrect parameter value k"
        assert (
            False == layers[-1].custom.parameters["transpose_y"].boolValue
        ), "Incorrect parameter value k"
        assert (
            True == layers[-1].custom.parameters["x_is_sparse"].boolValue
        ), "Incorrect parameter value k"
        assert (
            True == layers[-1].custom.parameters["y_is_sparse"].boolValue
        ), "Incorrect parameter value k"



SPACES = "  "

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from .block import curr_block, Block, Function
from .input_type import *
from .operation import *
from .program import *
from .var import *

from .builder import Builder
from .ops.defs._op_reqs import register_op



# Copyright (c) 2017, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import pandas as pd
import numpy as np
import random
import tempfile
import unittest
import pytest

from coremltools._deps import (
    _HAS_LIBSVM,
    MSG_LIBSVM_NOT_FOUND,
    _HAS_SKLEARN,
    MSG_SKLEARN_NOT_FOUND,
)
from coremltools.models.utils import evaluate_regressor, _macos_version, _is_macos

if _HAS_LIBSVM:
    import svmutil
    import svm
    from coremltools.converters import libsvm

if _HAS_SKLEARN:
    from sklearn.svm import SVR
    from sklearn.datasets import load_boston
    from coremltools.converters import sklearn as sklearn_converter
    from sklearn.preprocessing import OneHotEncoder


@unittest.skipIf(not _HAS_SKLEARN, MSG_SKLEARN_NOT_FOUND)
class SvrScikitTest(unittest.TestCase):
    """
    Unit test class for testing scikit-learn sklearn_converter.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        if not _HAS_SKLEARN:
            return

        scikit_data = load_boston()
        scikit_model = SVR(kernel="linear")
        scikit_model.fit(scikit_data["data"], scikit_data["target"])

        # Save the data and the model
        self.scikit_data = scikit_data
        self.scikit_model = scikit_model

    def test_conversion_bad_inputs(self):
        # Error on converting an untrained model
        with self.assertRaises(TypeError):
            model = SVR()
            spec = sklearn_converter.convert(model, "data", "out")

        # Check the expected class during covnersion.
        with self.assertRaises(TypeError):
            model = OneHotEncoder()
            spec = sklearn_converter.convert(model, "data", "out")

    @pytest.mark.slow
    def test_evaluation_stress_test(self):
        self._test_evaluation(allow_slow=True)

    def test_evaluation(self):
        self._test_evaluation(allow_slow=False)

    def _test_evaluation(self, allow_slow):
        """
        Test that the same predictions are made
        """

        # Generate some smallish (some kernels take too long on anything else) random data
        x, y = [], []
        for _ in range(50):
            cur_x1, cur_x2 = random.gauss(2, 3), random.gauss(-1, 2)
            x.append([cur_x1, cur_x2])
            y.append(1 + 2 * cur_x1 + 3 * cur_x2)

        input_names = ["x1", "x2"]
        df = pd.DataFrame(x, columns=input_names)

        # Parameters to test
        kernel_parameters = [
            {},
            {"kernel": "rbf", "gamma": 1.2},
            {"kernel": "linear"},
            {"kernel": "poly"},
            {"kernel": "poly", "degree": 2},
            {"kernel": "poly", "gamma": 0.75},
            {"kernel": "poly", "degree": 0, "gamma": 0.9, "coef0": 2},
            {"kernel": "sigmoid"},
            {"kernel": "sigmoid", "gamma": 1.3},
            {"kernel": "sigmoid", "coef0": 0.8},
            {"kernel": "sigmoid", "coef0": 0.8, "gamma": 0.5},
        ]
        non_kernel_parameters = [
            {},
            {"C": 1},
            {"C": 1.5, "epsilon": 0.5, "shrinking": True},
            {"C": 0.5, "epsilon": 1.5, "shrinking": False},
        ]

        # Test
        for param1 in non_kernel_parameters:
            for param2 in kernel_parameters:
                cur_params = param1.copy()
                cur_params.update(param2)
                print("cur_params=" + str(cur_params))

                cur_model = SVR(**cur_params)
                cur_model.fit(x, y)
                df["prediction"] = cur_model.predict(x)

                spec = sklearn_converter.convert(cur_model, input_names, "target")

                if _is_macos() and _macos_version() >= (10, 13):
                    metrics = evaluate_regressor(spec, df)
                    self.assertAlmostEquals(metrics["max_error"], 0)

                if not allow_slow:
                    break

            if not allow_slow:
                break


@unittest.skipIf(not _HAS_LIBSVM, MSG_LIBSVM_NOT_FOUND)
@unittest.skipIf(not _HAS_SKLEARN, MSG_SKLEARN_NOT_FOUND)
class EpsilonSVRLibSVMTest(unittest.TestCase):
    """
    Unit test class for testing the libsvm sklearn converter.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        if not _HAS_SKLEARN:
            return
        if not _HAS_LIBSVM:
            return

        scikit_data = load_boston()
        prob = svmutil.svm_problem(scikit_data["target"], scikit_data["data"].tolist())
        param = svmutil.svm_parameter()
        param.svm_type = svmutil.EPSILON_SVR
        param.kernel_type = svmutil.LINEAR
        param.eps = 1

        self.libsvm_model = svmutil.svm_train(prob, param)

    def test_input_names(self):
        data = load_boston()
        df = pd.DataFrame({"input": data["data"].tolist()})
        df["input"] = df["input"].apply(np.array)

        # Default values
        spec = libsvm.convert(self.libsvm_model)
        if _is_macos() and _macos_version() >= (10, 13):
            (df["prediction"], _, _) = svmutil.svm_predict(
                data["target"], data["data"].tolist(), self.libsvm_model
            )
            metrics = evaluate_regressor(spec, df)
            self.assertAlmostEquals(metrics["max_error"], 0)

        # One extra parameters. This is legal/possible.
        num_inputs = len(data["data"][0])
        spec = libsvm.convert(self.libsvm_model, input_length=num_inputs + 1)

        # Not enought input names.
        input_names = ["this", "is", "not", "enought", "names"]
        with self.assertRaises(ValueError):
            libsvm.convert(self.libsvm_model, input_names=input_names)
        with self.assertRaises(ValueError):
            libsvm.convert(self.libsvm_model, input_length=num_inputs - 1)

    def test_conversion_from_filesystem(self):
        libsvm_model_path = tempfile.mktemp(suffix="model.libsvm")
        svmutil.svm_save_model(libsvm_model_path, self.libsvm_model)
        spec = libsvm.convert(
            libsvm_model_path, input_names="data", target_name="target"
        )

    def test_conversion_bad_inputs(self):
        # Check the expected class during covnersion.
        with self.assertRaises(TypeError):
            model = OneHotEncoder()
            spec = libsvm.convert(model, "data", "out")

    @pytest.mark.slow
    def test_evaluation_stress_test(self):
        self._test_evaluation(allow_slow=True)

    def test_evaluation(self):
        self._test_evaluation(allow_slow=False)

    def _test_evaluation(self, allow_slow):
        """
        Test that the same predictions are made
        """
        from svm import svm_parameter, svm_problem
        from svmutil import svm_train, svm_predict

        # Generate some smallish (poly kernels take too long on anything else) random data
        x, y = [], []
        for _ in range(50):
            cur_x1, cur_x2 = random.gauss(2, 3), random.gauss(-1, 2)
            x.append([cur_x1, cur_x2])
            y.append(1 + 2 * cur_x1 + 3 * cur_x2)

        input_names = ["x1", "x2"]
        df = pd.DataFrame(x, columns=input_names)
        prob = svm_problem(y, x)

        # Parameters
        base_param = "-s 3"  # model type is epsilon SVR
        non_kernel_parameters = ["", "-c 1.5 -p 0.5 -h 1", "-c 0.5 -p 0.5 -h 0"]
        kernel_parameters = [
            "",
            "-t 2 -g 1.2",  # rbf kernel
            "-t 0",  # linear kernel
            "-t 1",
            "-t 1 -d 2",
            "-t 1 -g 0.75",
            "-t 1 -d 0 -g 0.9 -r 2",  # poly kernel
            "-t 3",
            "-t 3 -g 1.3",
            "-t 3 -r 0.8",
            "-t 3 -r 0.8 -g 0.5",  # sigmoid kernel
        ]

        for param1 in non_kernel_parameters:
            for param2 in kernel_parameters:
                param_str = " ".join([base_param, param1, param2])
                print(param_str)
                param = svm_parameter(param_str)

                model = svm_train(prob, param)
                (df["prediction"], _, _) = svm_predict(y, x, model)

                spec = libsvm.convert(
                    model, input_names=input_names, target_name="target"
                )

                if _is_macos() and _macos_version() >= (10, 13):
                    metrics = evaluate_regressor(spec, df)
                    self.assertAlmostEquals(metrics["max_error"], 0)

                if not allow_slow:
                    break

            if not allow_slow:
                break



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools import ImageType, models
from coremltools.converters.mil.testing_reqs import ct
from coremltools.converters.mil.testing_utils import (
    assert_op_count_match,
    assert_model_is_valid,
    get_op_types_in_program,
    apply_pass_and_basic_check,
)
from coremltools.converters.mil.mil import Builder as mb
import unittest

class ImagePreprocessingPass(unittest.TestCase):
    """
    Input graph:
    input (format=NHWC) ------> transpose(axis=[0, 3, 1, 2]) ---------> add ----> relu ---> out
                           |                                             ^
                           |                                             |
                           ---> relu ---> transpose(axis=[0, 3, 1, 2]) ---

    Intermediate graph:
    input (format=NCHW) -----> transpose(axis=[0, 2, 3, 1]) ----> transpose(axis=[0, 3, 1, 2]) ---------> add ----> relu ---> out
                                                              |                                             ^
                                                              |                                             |
                                                              ---> relu ---> transpose(axis=[0, 3, 1, 2]) ---


    Output graph:
    input (format=NCHW) -----> relu -----> add -----> relu -----> out
                          |                 ^
                          |                 |
                          -------------------
    """

    def test_fusion_with_image_intermediate_graph(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 20, 30, 3))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x2 = mb.relu(x=x)
            x3 = mb.transpose(x=x2, perm=[0, 3, 1, 2])
            x4 = mb.add(x=x1, y=x3)
            return mb.relu(x=x4)

        prog.main_input_types = [ImageType(name="x", shape=(10, 20, 30, 3), channel_first=False)]
        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::image_input_preprocess"
        )
        self.assertEqual(get_op_types_in_program(prev_prog), ["transpose", "relu", "transpose", "add", "relu"])
        self.assertEqual(get_op_types_in_program(prog), ["transpose", "transpose", "relu", "transpose", "add", "relu"])

    def test_fusion_with_image_full(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 20, 30, 3))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x2 = mb.relu(x=x)
            x3 = mb.transpose(x=x2, perm=[0, 3, 1, 2])
            x4 = mb.add(x=x1, y=x3)
            return mb.relu(x=x4)

        mlmodel = ct.convert(prog,
            inputs=[ImageType(name="x", shape=(10, 20, 30, 3),
              channel_first=False)],
            source="mil", convert_to="nn_proto")
        assert mlmodel is not None
        assert len(mlmodel.get_spec().neuralNetwork.layers) == 3



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.mil import get_new_symbol
from coremltools.converters.mil.testing_reqs import *

from .testing_utils import run_compare_builder

backends = testing_reqs.backends


class TestGRU:
    @pytest.mark.skipif(not testing_reqs._HAS_TORCH, reason="PyTorch not installed.")
    @pytest.mark.parametrize(
        argnames=[
            "use_cpu_only",
            "backend",
            "seq_len",
            "batch_size",
            "input_size",
            "hidden_size",
            "has_bias",
            "output_sequence",
            "direction",
            "symbolic",
        ],
        argvalues=itertools.product(
            [True, False],
            backends,
            [1, 8],
            [
                1
            ],  # <rdar://problem/59644603> [MIL] GRU with batch size 1 produces incorrect
            # output(always 0) for second batch onwards
            [2, 32],
            [1, 16],
            # rdar://66661491 (GRU with bias fails on NNv1 and MIL backend)
            [False],
            [True, False],
            ["forward", "reverse"],
            [True, False],
        ),
    )
    def test_builder_to_backend_smoke(
        self,
        use_cpu_only,
        backend,
        seq_len,
        batch_size,
        input_size,
        hidden_size,
        has_bias,
        output_sequence,
        direction,
        symbolic,
    ):
        torch.manual_seed(5)
        rnn = torch.nn.GRU(input_size, hidden_size, 1, bias=has_bias)
        state_dict = rnn.state_dict()

        state_dict["weight_ih_l0"] /= state_dict["weight_ih_l0"]
        state_dict["weight_hh_l0"] /= state_dict["weight_hh_l0"]

        ih_wt = state_dict["weight_ih_l0"].detach().numpy()
        hh_wt = state_dict["weight_hh_l0"].detach().numpy()

        # Make weight compatible to CoreML format
        def rzo_to_zro(x):
            r, z, o = np.split(x, 3)
            return np.concatenate([z, r, o], axis=0)

        w = np.concatenate([ih_wt, hh_wt], axis=1)
        w = rzo_to_zro(w).transpose()

        b = None
        if has_bias:
            ih_b = state_dict["bias_ih_l0"].detach().numpy()
            hh_b = state_dict["bias_hh_l0"].detach().numpy()
            ih_b = rzo_to_zro(ih_b)
            hh_b = rzo_to_zro(hh_b)
            b = np.stack([ih_b, hh_b], axis=0)

        t = torch.randn(seq_len, batch_size, input_size)
        h0 = torch.randn(1, batch_size, hidden_size)

        n_t = t
        if direction == "reverse":
            n_t = torch.flip(n_t, [0])

        output, hn = rnn(n_t, h0)
        if output_sequence == False:
            output = output[-1].unsqueeze(0)

        output = output.detach().numpy()
        hn = hn.detach().numpy().squeeze(0)

        t = np.reshape(t.detach().numpy(), [seq_len, batch_size, input_size])
        h = np.reshape(h0.detach().numpy().squeeze(0), [batch_size, hidden_size])

        if symbolic:
            batch_size = get_new_symbol()
            seq_len = get_new_symbol()

        input_shape = [seq_len, batch_size, input_size]
        h_shape = [batch_size, hidden_size]

        expected_output_types = [
            (seq_len if output_sequence else 1, batch_size, hidden_size, types.fp32),
            (batch_size, hidden_size, types.fp32),
        ]
        expected_outputs = [output, hn]

        input_placeholders = {
            "x": mb.placeholder(shape=input_shape),
            "initial_h": mb.placeholder(shape=h_shape),
        }
        input_values = {"x": t, "initial_h": h}

        def build(x, initial_h):
            arguments = {
                "x": x,
                "initial_h": initial_h,
                "weight": w,
                "direction": direction,
                "output_sequence": output_sequence,
            }
            # If bias is provided, add in arguments
            if b is not None:
                arguments["bias"] = b
            return mb.gru(**arguments)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )


class TestLSTM:
    @pytest.mark.parametrize(
        ",".join(
            [
                "use_cpu_only",
                "backend",
                "input_dims",
                "output_dim",
                "activation",
                "inner_activation",
                "outer_activation",
                "return_seq",
                "has_bias",
                "forget_bias",
                "has_peephole",
                "coupled_input_forget",
                "clip",
            ]
        ),
        itertools.product(
            [True, False],
            backends,
            [[8, 32, 32]],
            [1, 4],
            ["SIGMOID"],
            ["TANH"],
            ["TANH", "SIGMOID"],
            [False, True],
            [False, True],
            [False, True],
            [True, False],
            [False],  # We have not exposed this option yet!
            [50.0, 0.2, 0.01],
        ),
    )
    def test_numpy_numerical(
        self,
        use_cpu_only,
        backend,
        input_dims,
        output_dim,
        activation,
        inner_activation,
        outer_activation,
        return_seq,
        has_bias,
        forget_bias,
        has_peephole,
        coupled_input_forget,
        clip,
    ):
        def _apply_act(x, option):
            if option == "TANH":
                return np.tanh(x)
            elif option == "RELU":
                return np.maximum(0, x)
            elif option == "SIGMOID":
                return 1.0 / (1 + np.exp(-x))
            elif option == "SIGMOID_HARD":
                return np.minimum(np.maximum(0.2 * x + 0.5, 0), 1)
            elif option == "LINEAR":
                return x
            else:
                raise ValueError("activation invalid")

        def _clip(x, threshold=500.0):
            return np.maximum(np.minimum(x, threshold), -threshold)

        def _get_numpy_prediction_lstm(Weights, X):
            # X : (batch, seq_len, channel)
            batch, _, _ = X.shape
            out = []
            for i in range(batch):
                out.append(
                    _get_numpy_prediction_lstm_single_batch(
                        Weights, np.expand_dims(X[i, :, :], axis=0)
                    )
                )
            return np.stack(out, axis=0)

        def _get_numpy_prediction_lstm_single_batch(Weights, X):

            batch_size, seq_len, input_size = X.shape
            X = X[0, :, :]
            hidden_size = output_dim

            b = Weights["b"]
            b = b[0] + b[1]
            Wx_i, Wx_f, Wx_o, Wx_g = np.split(Weights["W_x"], 4)
            Wh_i, Wh_f, Wh_o, Wh_g = np.split(Weights["W_h"], 4)
            b_i, b_f, b_o, b_g = np.split(b, 4)
            p_i, p_f, p_o = np.split(Weights["p"], 3)

            act1 = activation
            act2 = inner_activation
            act3 = outer_activation

            h = np.zeros((hidden_size))
            c = np.zeros((hidden_size))
            np_out = np.zeros((seq_len, hidden_size))
            for k in range(seq_len):
                x = X[k, :]
                i = _apply_act(np.dot(Wx_i, x) + np.dot(Wh_i, h) + b_i + c * p_i, act1)
                f = _apply_act(np.dot(Wx_f, x) + np.dot(Wh_f, h) + b_f + c * p_f, act1)
                g = _apply_act(np.dot(Wx_g, x) + np.dot(Wh_g, h) + b_g, act2)
                if coupled_input_forget:
                    c = c * (1 - i) + i * g
                else:
                    c = c * f + i * g
                c = _clip(c, clip)
                o = _apply_act(np.dot(Wx_o, x) + np.dot(Wh_o, h) + b_o + c * p_o, act1)
                h = o * _apply_act(c, act3)
                np_out[k, :] = h

            if return_seq:
                np_out_final = np_out
            else:
                np_out_final = np_out[-1, :]
            return np_out_final

        batch = input_dims[0]
        seq_len = input_dims[1]
        input_size = input_dims[2]
        hidden_size = output_dim

        # define random weights
        W_x = np.random.rand(4 * hidden_size, input_size)
        W_h = np.random.rand(4 * hidden_size, hidden_size)

        if has_bias:
            b = np.random.rand(2, 4 * hidden_size) - 0.5
            if forget_bias:
                b = b + 1
        else:
            b = np.zeros((2, 4 * hidden_size))

        if has_peephole:
            p = np.random.rand(3 * hidden_size) - 0.5
        else:
            p = np.zeros((3 * hidden_size))

        Weights = {}
        Weights["W_x"] = W_x
        Weights["W_h"] = W_h
        Weights["b"] = b
        Weights["p"] = p

        input_data = np.random.rand(batch, seq_len, input_size)
        numpy_preds = _get_numpy_prediction_lstm(Weights, input_data)
        if return_seq:
            numpy_preds = np.transpose(numpy_preds, [1, 0, 2])

        coreml_input_data = np.transpose(input_data, [1, 0, 2])
        input_placeholders = {"x": mb.placeholder(shape=coreml_input_data.shape)}
        input_values = {"x": coreml_input_data}

        weights = np.concatenate([W_x, W_h], axis=1).transpose()

        def build(x):
            h_all, ht, ct = mb.lstm(
                x=x,
                initial_h=np.zeros((batch, hidden_size)).astype(np.float32),
                initial_c=np.zeros((batch, hidden_size)).astype(np.float32),
                weight=weights,
                peephole=p,
                direction="forward",
                bias=b,
                output_sequence=return_seq,
                recurrent_activation=activation,
                cell_activation=inner_activation,
                activation=outer_activation,
                clip=clip,
            )
            return h_all

        expected_output_types = (
            seq_len if return_seq else 1,
            batch,
            hidden_size,
            types.fp32,
        )
        expected_outputs = numpy_preds

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
            # rdar://63839623 ([GITLAB-CI] precision issue on various tests on gitlab ci)
            atol=1e-3,
            rtol=1e-3,
        )

    @pytest.mark.skipif(not testing_reqs._HAS_TORCH, reason="PyTorch not installed.")
    @pytest.mark.parametrize(
        argnames=[
            "use_cpu_only",
            "backend",
            "seq_len",
            "batch_size",
            "input_size",
            "hidden_size",
            "has_bias",
            "output_sequence",
            "direction",
            "symbolic",
        ],
        argvalues=itertools.product(
            [True, False],
            backends,
            [1, 8],
            [1, 32],
            [1, 64],
            [1, 16],
            [True, False],
            [True, False],
            ["forward", "reverse"],
            [True, False],
        ),
    )
    def test_builder_to_backend_smoke_unilstm(
        self,
        use_cpu_only,
        backend,
        seq_len,
        batch_size,
        input_size,
        hidden_size,
        has_bias,
        output_sequence,
        direction,
        symbolic,
    ):
        # TODO: <rdar://problem/59540160> [MIL] LSTM layer- Implement eval and tf register routine
        # Testing 1. peephole values
        #         2. clip values

        torch.manual_seed(50)
        rnn = torch.nn.LSTM(input_size, hidden_size, 1, bias=has_bias)
        state_dict = rnn.state_dict()

        ih_wt = state_dict["weight_ih_l0"].detach().numpy()
        hh_wt = state_dict["weight_hh_l0"].detach().numpy()

        # Make weight compatible to CoreML format
        def ifzo_to_ifoz(x):
            i, f, z, o = np.split(x, 4)
            return np.concatenate([i, f, o, z], axis=0)

        w = np.concatenate([ih_wt, hh_wt], axis=1)
        w = ifzo_to_ifoz(w).transpose()

        b = None
        if has_bias:
            ih_b = state_dict["bias_ih_l0"].detach().numpy()
            hh_b = state_dict["bias_hh_l0"].detach().numpy()
            ih_b = ifzo_to_ifoz(ih_b).transpose()
            hh_b = ifzo_to_ifoz(hh_b).transpose()
            b = np.stack([ih_b, hh_b], axis=0)

        t = torch.randn(seq_len, batch_size, input_size)
        h0 = torch.randn(1, batch_size, hidden_size)
        c0 = torch.randn(1, batch_size, hidden_size)

        n_t = t
        if direction == "reverse":
            n_t = torch.flip(n_t, [0])

        output, (hn, cn) = rnn(n_t, (h0, c0))
        if output_sequence == False:
            output = output[-1].unsqueeze(0)

        output = output.detach().numpy()
        hn = hn.detach().numpy().squeeze(0)
        cn = cn.detach().numpy().squeeze(0)

        t = np.reshape(t.detach().numpy(), [seq_len, batch_size, input_size])
        h = np.reshape(h0.detach().numpy().squeeze(0), [batch_size, hidden_size])
        c = np.reshape(c0.detach().numpy().squeeze(0), [batch_size, hidden_size])

        if symbolic:
            batch_size = get_new_symbol()
            seq_len = get_new_symbol()

        input_shape = [seq_len, batch_size, input_size]
        h_shape = [batch_size, hidden_size]
        c_shape = [batch_size, hidden_size]

        expected_output_types = [
            (seq_len if output_sequence else 1, batch_size, hidden_size, types.fp32),
            (batch_size, hidden_size, types.fp32),
            (batch_size, hidden_size, types.fp32),
        ]
        expected_outputs = [output, hn, cn]

        input_placeholders = {
            "x": mb.placeholder(shape=input_shape),
            "initial_h": mb.placeholder(shape=h_shape),
            "initial_c": mb.placeholder(shape=c_shape),
        }
        input_values = {"x": t, "initial_h": h, "initial_c": c}

        def build(x, initial_h, initial_c):
            arguments = {
                "x": x,
                "initial_h": initial_h,
                "initial_c": initial_c,
                "weight": w,
                "direction": direction,
                "output_sequence": output_sequence,
            }
            # If bias is provided, add in arguments
            if b is not None:
                arguments["bias"] = b
            return mb.lstm(**arguments)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @pytest.mark.skipif(not testing_reqs._HAS_TORCH, reason="PyTorch not installed.")
    @pytest.mark.parametrize(
        argnames=[
            "use_cpu_only",
            "backend",
            "seq_len",
            "batch_size",
            "input_size",
            "hidden_size",
            "has_bias",
            "output_sequence",
            "symbolic",
        ],
        argvalues=itertools.product(
            [True],
            # TODO: rdar://66768742 (BiLSTM output numerically mismatch for MIL backend)
            ["nn_proto"],
            [1, 8],
            [1, 32],
            [1, 64],
            [2, 16],
            [True, False],
            [True, False],
            [True, False],
        ),
    )
    def test_builder_to_backend_smoke_bidirlstm(
        self,
        use_cpu_only,
        backend,
        seq_len,
        batch_size,
        input_size,
        hidden_size,
        has_bias,
        output_sequence,
        symbolic,
    ):
        def _pytorch_hidden_to_coreml(x):
            x = x.detach().numpy()
            # Split of Direction axis
            f, b = np.split(x, 2, axis=0)
            # Concat on Hidden Size axis
            x = np.concatenate([f, b], axis=2)
            x = np.squeeze(x, axis=0)
            return x

        direction = "bidirectional"
        torch.manual_seed(20)
        rnn = torch.nn.LSTM(
            input_size, hidden_size, 1, bidirectional=True, bias=has_bias
        )
        state_dict = rnn.state_dict()

        ih_wt = state_dict["weight_ih_l0"].detach().numpy()
        hh_wt = state_dict["weight_hh_l0"].detach().numpy()
        ih_wt_r = state_dict["weight_ih_l0_reverse"].detach().numpy()
        hh_wt_r = state_dict["weight_hh_l0_reverse"].detach().numpy()

        f_wt = np.concatenate([ih_wt, hh_wt], axis=1)
        r_wt = np.concatenate([ih_wt_r, hh_wt_r], axis=1)

        def ifzo_to_ifoz(x):
            i, f, z, o = np.split(x, 4)
            return np.concatenate([i, f, o, z], axis=0)

        f_wt = ifzo_to_ifoz(f_wt).transpose()
        r_wt = ifzo_to_ifoz(r_wt).transpose()
        w = np.concatenate([f_wt, r_wt], axis=1)

        b = None
        if has_bias:
            ih_b = state_dict["bias_ih_l0"].detach().numpy()
            hh_b = state_dict["bias_hh_l0"].detach().numpy()
            ih_b_r = state_dict["bias_ih_l0_reverse"].detach().numpy()
            hh_b_r = state_dict["bias_hh_l0_reverse"].detach().numpy()
            # Convert forward bias into [2, 4*H]
            ih_b = ifzo_to_ifoz(ih_b)
            hh_b = ifzo_to_ifoz(hh_b)
            f_b = np.stack([ih_b, hh_b], axis=0)
            # Convert reverse bias into [2, 4*H]
            ih_b_r = ifzo_to_ifoz(ih_b_r)
            hh_b_r = ifzo_to_ifoz(hh_b_r)
            r_b = np.stack([ih_b_r, hh_b_r], axis=0)
            # Final bias of [2, 2*4*H]
            b = np.concatenate([f_b, r_b], axis=1)

        t = torch.randn(seq_len, batch_size, input_size)
        h0 = torch.randn(2, batch_size, hidden_size)
        c0 = torch.randn(2, batch_size, hidden_size)

        output, (hn, cn) = rnn(t, (h0, c0))
        if output_sequence == False:
            output_f = output[-1].unsqueeze(0)[:, :, :hidden_size]
            output_r = output[0].unsqueeze(0)[:, :, hidden_size:]
            output = torch.cat([output_f, output_r], dim=2)

        output = output.detach().numpy()
        hn = _pytorch_hidden_to_coreml(hn)
        cn = _pytorch_hidden_to_coreml(cn)

        if symbolic:
            batch_size = get_new_symbol()
            seq_len = get_new_symbol()

        input_shape = [seq_len, batch_size, input_size]
        h_shape = [batch_size, 2 * hidden_size]
        c_shape = [batch_size, 2 * hidden_size]

        expected_output_types = [
            (
                seq_len if output_sequence else 1,
                batch_size,
                2 * hidden_size,
                types.fp32,
            ),
            (batch_size, 2 * hidden_size, types.fp32),
            (batch_size, 2 * hidden_size, types.fp32),
        ]
        expected_outputs = [output, hn, cn]

        t = t.detach().numpy()
        h = _pytorch_hidden_to_coreml(h0)
        c = _pytorch_hidden_to_coreml(c0)

        input_placeholders = {
            "x": mb.placeholder(shape=input_shape),
            "initial_h": mb.placeholder(shape=h_shape),
            "initial_c": mb.placeholder(shape=c_shape),
        }
        input_values = {"x": t, "initial_h": h, "initial_c": c}

        def build(x, initial_h, initial_c):
            arguments = {
                "x": x,
                "initial_h": initial_h,
                "initial_c": initial_c,
                "weight": w,
                "direction": direction,
                "output_sequence": output_sequence,
            }
            # If bias is provided, add in arguments
            if b is not None:
                arguments["bias"] = b
            return mb.lstm(**arguments)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )


class TestRNN:
    @pytest.mark.skipif(not testing_reqs._HAS_TORCH, reason="PyTorch not installed.")
    @pytest.mark.parametrize(
        argnames=[
            "use_cpu_only",
            "backend",
            "seq_len",
            "batch_size",
            "input_size",
            "hidden_size",
            "has_bias",
            "output_sequence",
            "direction",
            "symbolic",
        ],
        argvalues=itertools.product(
            [True, False],
            backends,
            [2, 8],
            [1, 32],
            [1, 64],
            [1, 16],
            [True, False],
            [True, False],
            ["forward", "reverse"],
            [True, False],
        ),
    )
    def test_builder_to_backend_smoke(
        self,
        use_cpu_only,
        backend,
        seq_len,
        batch_size,
        input_size,
        hidden_size,
        has_bias,
        output_sequence,
        direction,
        symbolic,
    ):
        torch.manual_seed(50)
        rnn = torch.nn.RNN(input_size, hidden_size, 1, bias=has_bias)
        state_dict = rnn.state_dict()

        ih_wt = state_dict["weight_ih_l0"].detach().numpy()
        hh_wt = state_dict["weight_hh_l0"].detach().numpy()
        w = np.concatenate([ih_wt, hh_wt], axis=1).transpose()

        b = None
        if has_bias:
            ih_b = state_dict["bias_ih_l0"].detach().numpy()
            hh_b = state_dict["bias_hh_l0"].detach().numpy()
            b = np.stack([ih_b, hh_b], axis=0)

        t = torch.randn(seq_len, batch_size, input_size)
        h0 = torch.randn(1, batch_size, hidden_size)

        n_t = t
        if direction == "reverse":
            n_t = torch.flip(n_t, [0])

        output, hn = rnn(n_t, h0)
        if output_sequence == False:
            output = output[-1].unsqueeze(0)

        output = output.detach().numpy()
        hn = hn.detach().numpy().squeeze(0)

        t = np.reshape(t.detach().numpy(), [seq_len, batch_size, input_size])
        h = np.reshape(h0.detach().numpy().squeeze(0), [batch_size, hidden_size])

        if symbolic:
            batch_size = get_new_symbol()
            seq_len = get_new_symbol()

        input_shape = [seq_len, batch_size, input_size]
        h_shape = [batch_size, hidden_size]

        expected_output_types = [
            (seq_len if output_sequence else 1, batch_size, hidden_size, types.fp32),
            (batch_size, hidden_size, types.fp32),
        ]
        expected_outputs = [output, hn]

        input_placeholders = {
            "x": mb.placeholder(shape=input_shape),
            "initial_h": mb.placeholder(shape=h_shape),
        }
        input_values = {"x": t, "initial_h": h}

        def build(x, initial_h):
            arguments = {
                "x": x,
                "initial_h": initial_h,
                "weight": w,
                "direction": direction,
                "output_sequence": output_sequence,
            }
            # If bias is provided, add in arguments
            if b is not None:
                arguments["bias"] = b
            return mb.rnn(**arguments)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )



# Copyright (c) 2019, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import unittest

import numpy as np
from scipy import sparse

from coremltools._deps import _HAS_SKLEARN

if _HAS_SKLEARN:
    from coremltools.converters import sklearn
    from sklearn.datasets import load_iris
    from sklearn.neighbors import KNeighborsClassifier


@unittest.skipIf(not _HAS_SKLEARN, "Missing sklearn. Skipping tests.")
class KNeighborsClassifierScikitTest(unittest.TestCase):
    """
    Unit test class for testing scikit-learn converter.
    """

    @classmethod
    def setUpClass(self):
        """
        Set up the unit test by loading the dataset and training a model.
        """
        print("Setting up KNeighborsClassifier converter tests")
        iris_samples = load_iris()
        self.iris_X = iris_samples.data
        self.iris_y = iris_samples.target

    def test_conversion_unfitted(self):
        """Tests conversion failure for an unfitted scikit model."""
        scikit_model = KNeighborsClassifier()
        self.assertRaises(TypeError, sklearn.convert, scikit_model)

    def test_conversion_brute_algorithm(self):
        """Tests conversion of a scikit KNeighborsClassifier using the brute force algorithm."""
        scikit_model = KNeighborsClassifier(algorithm="brute", n_neighbors=42)
        scikit_model.fit(self.iris_X, self.iris_y)

        coreml_model = sklearn.convert(scikit_model, "single_input", "single_output")
        coreml_spec = coreml_model.get_spec()

        self.assertIsNotNone(coreml_spec)
        self.assertTrue(coreml_spec.HasField("kNearestNeighborsClassifier"))
        self.assertEqual(
            coreml_spec.kNearestNeighborsClassifier.numberOfNeighbors.defaultValue, 42
        )
        self.assertEqual(
            coreml_spec.kNearestNeighborsClassifier.numberOfNeighbors.range.minValue, 1
        )
        self.assertEqual(
            coreml_spec.kNearestNeighborsClassifier.numberOfNeighbors.range.maxValue,
            len(self.iris_X),
        )
        self.assertTrue(
            coreml_spec.kNearestNeighborsClassifier.HasField("uniformWeighting")
        )
        self.assertEqual(
            coreml_spec.kNearestNeighborsClassifier.nearestNeighborsIndex.numberOfDimensions,
            len(self.iris_X[0]),
        )
        self.assertTrue(
            coreml_spec.kNearestNeighborsClassifier.nearestNeighborsIndex.HasField(
                "linearIndex"
            )
        )
        self.assertTrue(
            coreml_spec.kNearestNeighborsClassifier.nearestNeighborsIndex.HasField(
                "squaredEuclideanDistance"
            )
        )

        self.validate_labels(coreml_spec, self.iris_y)
        self.validate_float_samples(coreml_spec, self.iris_X)

    def test_conversion_kd_tree_algorithm(self):
        """Tests conversion of a scikit KNeighborsClassifier using the brute force algorithm."""
        test_leaf_size = 23
        test_n_neighbors = 42
        scikit_model = KNeighborsClassifier(
            algorithm="kd_tree", leaf_size=test_leaf_size, n_neighbors=test_n_neighbors
        )
        scikit_model.fit(self.iris_X, self.iris_y)

        coreml_model = sklearn.convert(scikit_model, "single_input", "single_output")
        coreml_spec = coreml_model.get_spec()

        self.assertIsNotNone(coreml_spec)
        self.assertTrue(coreml_spec.HasField("kNearestNeighborsClassifier"))
        self.assertEqual(
            coreml_spec.kNearestNeighborsClassifier.numberOfNeighbors.defaultValue,
            test_n_neighbors,
        )
        self.assertEqual(
            coreml_spec.kNearestNeighborsClassifier.numberOfNeighbors.range.minValue, 1
        )
        self.assertEqual(
            coreml_spec.kNearestNeighborsClassifier.numberOfNeighbors.range.maxValue,
            len(self.iris_X),
        )
        self.assertTrue(
            coreml_spec.kNearestNeighborsClassifier.HasField("uniformWeighting")
        )
        self.assertEqual(
            coreml_spec.kNearestNeighborsClassifier.nearestNeighborsIndex.numberOfDimensions,
            len(self.iris_X[0]),
        )
        self.assertTrue(
            coreml_spec.kNearestNeighborsClassifier.nearestNeighborsIndex.HasField(
                "singleKdTreeIndex"
            )
        )
        self.assertEqual(
            test_leaf_size,
            coreml_spec.kNearestNeighborsClassifier.nearestNeighborsIndex.singleKdTreeIndex.leafSize,
        )
        self.assertTrue(
            coreml_spec.kNearestNeighborsClassifier.nearestNeighborsIndex.HasField(
                "squaredEuclideanDistance"
            )
        )

        self.validate_labels(coreml_spec, self.iris_y)
        self.validate_float_samples(coreml_spec, self.iris_X)

    def test_conversion_auto_algorithm(self):
        """Tests conversion of a scikit KNeighborsClassifier using the brute force algorithm."""
        test_n_neighbors = 42
        scikit_model = KNeighborsClassifier(
            algorithm="auto", n_neighbors=test_n_neighbors
        )
        scikit_model.fit(self.iris_X, self.iris_y)

        coreml_model = sklearn.convert(scikit_model, "single_input", "single_output")
        coreml_spec = coreml_model.get_spec()
        self.assertIsNotNone(coreml_spec)

    def test_conversion_unsupported_algorithm(self):
        """Test a scikit KNeighborsClassifier with an invalid algorithm."""
        scikit_model = KNeighborsClassifier(algorithm="ball_tree")
        self.assertRaises(TypeError, sklearn.convert, scikit_model)

    def test_conversion_weight_function_good(self):
        scikit_model = KNeighborsClassifier(weights="uniform")
        scikit_model.fit(self.iris_X, self.iris_y)

        coreml_model = sklearn.convert(scikit_model, "single_input", "single_output")
        coreml_spec = coreml_model.get_spec()
        self.assertIsNotNone(coreml_spec)
        self.assertTrue(
            coreml_spec.kNearestNeighborsClassifier.HasField("uniformWeighting")
        )

    def test_conversion_unsupported_weight_function(self):
        scikit_model = KNeighborsClassifier(algorithm="brute", weights="distance")
        scikit_model.fit(self.iris_X, self.iris_y)
        self.assertRaises(TypeError, sklearn.convert, scikit_model)

        def callable_weight_function():
            print("Inside callable_weight_function")

        scikit_model = KNeighborsClassifier(
            algorithm="brute", weights=callable_weight_function
        )
        scikit_model.fit(self.iris_X, self.iris_y)
        self.assertRaises(TypeError, sklearn.convert, scikit_model)

    def test_conversion_distance_function_good(self):
        """Tests conversion of a scikit KNeighborsClassifier with a valid distance metric."""
        scikit_model = KNeighborsClassifier(algorithm="brute", metric="euclidean")
        scikit_model.fit(self.iris_X, self.iris_y)
        coreml_model = sklearn.convert(scikit_model, "single_input", "single_output")
        coreml_spec = coreml_model.get_spec()
        self.assertIsNotNone(coreml_spec)
        self.assertTrue(
            coreml_spec.kNearestNeighborsClassifier.nearestNeighborsIndex.HasField(
                "squaredEuclideanDistance"
            )
        )

        # Minkowski metric with p=2 is equivalent to the squared Euclidean distance
        scikit_model = KNeighborsClassifier(algorithm="brute", metric="minkowski", p=2)
        scikit_model.fit(self.iris_X, self.iris_y)
        coreml_spec = coreml_model.get_spec()
        self.assertIsNotNone(coreml_spec)
        self.assertTrue(
            coreml_spec.kNearestNeighborsClassifier.nearestNeighborsIndex.HasField(
                "squaredEuclideanDistance"
            )
        )

    def test_conversion_unsupported_distance_function(self):
        """Tests conversion of a scikit KNeighborsClassifier with an invalid distance metric."""
        # There are many possible distance functions for a brute force neighbors function, but these 3 should give us
        # coverage over the converter code.
        scikit_model = KNeighborsClassifier(algorithm="brute", metric="manhattan")
        scikit_model.fit(self.iris_X, self.iris_y)
        self.assertRaises(TypeError, sklearn.convert, scikit_model)

        scikit_model = KNeighborsClassifier(algorithm="kd_tree", metric="chebyshev")
        scikit_model.fit(self.iris_X, self.iris_y)
        self.assertRaises(TypeError, sklearn.convert, scikit_model)

        scikit_model = KNeighborsClassifier(algorithm="brute", metric="minkowski", p=3)
        scikit_model.fit(self.iris_X, self.iris_y)
        self.assertRaises(TypeError, sklearn.convert, scikit_model)

        def callable_distance_function():
            print("Inside callable_distance_function")

        scikit_model = KNeighborsClassifier(
            algorithm="brute", metric=callable_distance_function
        )
        scikit_model.fit(self.iris_X, self.iris_y)
        self.assertRaises(TypeError, sklearn.convert, scikit_model)

    def test_conversion_with_sparse_X(self):
        """Tests conversion of a model that's fitted with sparse data."""
        num_samples = 100
        num_dims = 64
        sparse_X = sparse.rand(
            num_samples, num_dims, format="csr"
        )  # KNeighborsClassifier only supports CSR format
        y = self.iris_y[
            0:num_samples
        ]  # the labels themselves don't matter - just use 100 of the Iris ones

        sklearn_model = KNeighborsClassifier(algorithm="brute")
        sklearn_model.fit(sparse_X, y)

        coreml_model = sklearn.convert(sklearn_model)
        coreml_spec = coreml_model.get_spec()
        self.assertIsNotNone(coreml_spec)

    def test_conversion_with_sparse_y(self):
        """Tests conversion of a model that's fitted with y values in a sparse format."""
        from sklearn.model_selection import train_test_split

        X_train, X_test, y_train, y_test = train_test_split(
            self.iris_X, self.iris_y, test_size=0.2, train_size=0.8
        )

        from sklearn import preprocessing

        lb = preprocessing.LabelBinarizer(sparse_output=True)
        binarized_y = lb.fit_transform(y_train)

        sklearn_model = KNeighborsClassifier(algorithm="brute")
        sklearn_model.fit(X_train, binarized_y)

        self.assertRaises(ValueError, sklearn.convert, sklearn_model)

    def validate_labels(self, spec, expected):
        """Validate the labels returned from the converted scikit KNeighborsClassifier"""
        self.assertTrue(spec.kNearestNeighborsClassifier.HasField("int64ClassLabels"))
        for index, label in enumerate(
            spec.kNearestNeighborsClassifier.int64ClassLabels.vector
        ):
            self.assertEqual(label, expected[index])

    def validate_float_samples(self, spec, expected):
        """Validate the float samples returned from the converted scikit KNeighborsClassifier"""
        num_dimensions = (
            spec.kNearestNeighborsClassifier.nearestNeighborsIndex.numberOfDimensions
        )
        for index, sample in enumerate(
            spec.kNearestNeighborsClassifier.nearestNeighborsIndex.floatSamples
        ):
            for dim in range(0, num_dimensions):
                self.assertAlmostEqual(
                    sample.vector[dim], expected[index][dim], places=6
                )



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb
import numpy as np


def _check_child_op_type(op, child_op_type):
    """
    :param op: operation
    :param child_op_type: str
    :return: Return True if op has 1 child and type of that child matches child_op_type
    """
    if len(op.outputs) != 1:
        return False
    child_ops = list(op.outputs[0].child_ops)
    if len(child_ops) != 1:
        return False
    if child_ops[0].op_type == child_op_type:
        return True
    return False


def try_to_transform(reduce_mean_op, block):
    all_ops = [reduce_mean_op]
    root_var = reduce_mean_op.x

    input_shape = root_var.shape
    if input_shape is None:
        return False

    rank = len(input_shape)

    # check that root_var feeds into exactly 3 ops
    if len(list(root_var.child_ops)) != 3:
        return False

    # check 1st reduce_mean op
    if not (
        reduce_mean_op.keep_dims.val is not None
        and reduce_mean_op.keep_dims.val == True
    ):
        return False
    axes = reduce_mean_op.axes.val
    if axes is None:
        return False

    # check 1st sub op
    child_ops_reduce_mean = list(reduce_mean_op.outputs[0].child_ops)
    if len(child_ops_reduce_mean) != 2:
        return False
    op_a = child_ops_reduce_mean[0]
    op_b = child_ops_reduce_mean[1]
    if not (
        (op_a.op_type == "sub" and op_b.op_type == "mul")
        or (op_a.op_type == "mul" and op_b.op_type == "sub")
    ):
        return False
    sub_op1 = op_a if op_a.op_type == "sub" else op_b
    if not (sub_op1.x == root_var and sub_op1.y == reduce_mean_op.outputs[0]):
        return False
    all_ops.append(sub_op1)

    # check square op
    if not _check_child_op_type(sub_op1, "square"):
        return False
    square_op = list(sub_op1.outputs[0].child_ops)[0]
    all_ops.append(square_op)

    # check second reduce mean
    if not _check_child_op_type(square_op, "reduce_mean"):
        return False
    reduce_mean_op2 = list(square_op.outputs[0].child_ops)[0]
    if not (
        reduce_mean_op2.keep_dims.val is not None
        and reduce_mean_op2.keep_dims.val == True
    ):
        return False
    if not (
        (reduce_mean_op2.axes.val is not None)
        and (axes == reduce_mean_op2.axes.val).all()
    ):
        return False
    all_ops.append(reduce_mean_op2)

    # check add op (with epsilon)
    if not _check_child_op_type(reduce_mean_op2, "add"):
        return False
    add_op1 = list(reduce_mean_op2.outputs[0].child_ops)[0]
    epsilon_var = add_op1.y if add_op1.x == reduce_mean_op2.outputs[0] else add_op1.x
    if epsilon_var.val is None:
        return False
    if len(epsilon_var.val.shape) != 0:
        # must be scalar
        return False
    all_ops.append(add_op1)

    # check rsqrt op
    if not _check_child_op_type(add_op1, "rsqrt"):
        return False
    rsqrt_op = list(add_op1.outputs[0].child_ops)[0]
    all_ops.append(rsqrt_op)

    # check mul (gamma)
    if not _check_child_op_type(rsqrt_op, "mul"):
        return False
    mul_op1 = list(rsqrt_op.outputs[0].child_ops)[0]
    gamma_var = mul_op1.y if mul_op1.x == rsqrt_op.outputs[0] else mul_op1.x
    if gamma_var.val is None:
        return False
    all_ops.append(mul_op1)

    # check 2 muls after the gamma mul
    child_ops = list(mul_op1.outputs[0].child_ops)
    if len(child_ops) != 2:
        return False
    mul_op2 = child_ops[0]
    mul_op3 = child_ops[1]
    if not (mul_op2.op_type == "mul" and mul_op3.op_type == "mul"):
        return False
    mul_op2_other_var = mul_op2.x if mul_op2.y == mul_op1.outputs[0] else mul_op2.y
    mul_op3_other_var = mul_op3.x if mul_op3.y == mul_op1.outputs[0] else mul_op3.y
    if not (
        (
            mul_op2_other_var == root_var
            and mul_op3_other_var == reduce_mean_op.outputs[0]
        )
        or (
            mul_op2_other_var == reduce_mean_op.outputs[0]
            and mul_op3_other_var == root_var
        )
    ):
        return False
    if mul_op2_other_var == root_var:
        mul_root_op = mul_op2
        mul_mean_op = mul_op3
    else:
        mul_root_op = mul_op3
        mul_mean_op = mul_op2
    all_ops.append(mul_mean_op)
    all_ops.append(mul_root_op)

    # check sub with beta
    if not _check_child_op_type(mul_mean_op, "sub"):
        return False
    sub_op2 = list(mul_mean_op.outputs[0].child_ops)[0]
    if sub_op2.y != mul_mean_op.outputs[0]:
        return False
    beta_var = sub_op2.x
    if beta_var.val is None:
        return False
    all_ops.append(sub_op2)

    # check last add op
    if not _check_child_op_type(sub_op2, "add"):
        return False
    add_op2 = list(sub_op2.outputs[0].child_ops)[0]
    if not (add_op2.x == mul_root_op.outputs[0] or add_op2.y == mul_root_op.outputs[0]):
        return False
    all_ops.append(add_op2)

    # check that none of the op in this pattern is connected to the output
    # (except the last add op)
    for i, op in enumerate(all_ops):
        if i == len(all_ops) - 1:
            continue
        for out in op.outputs:
            if out in block.outputs:
                return False

    # check whether the pattern is instance_norm or layer_norm
    is_layernorm = False
    is_instancenorm = False

    negative_axes = [a - rank if a >= 0 else a for a in axes]
    negative_axes.sort()

    if len(gamma_var.val.shape) == len(axes) and len(beta_var.val.shape) == len(axes):
        # axes for layer_norm must be [-1] or [-1,-2] or [-1,-2,-3] and so on
        if negative_axes == list(range(-len(negative_axes), 0)):
            is_layernorm = True

    if negative_axes == [-2, -1] and rank == 4:
        if (
            len(np.squeeze(gamma_var.val).shape) == 1
            and len(np.squeeze(beta_var.val).shape) == 1
        ):
            is_instancenorm = True

    if not (is_instancenorm or is_layernorm):
        return False

    # remove all the ops, and replace with a layer_norm or instance_norm op
    out_name = add_op2.outputs[0].name

    if is_instancenorm:
        x = mb.instance_norm(
            x=root_var,
            gamma=np.squeeze(gamma_var.val),
            beta=np.squeeze(beta_var.val),
            epsilon=epsilon_var,
            name=out_name,
            before_op=add_op2,
        )
    else:
        x = mb.layer_norm(
            x=root_var,
            axes=axes,
            gamma=gamma_var,
            beta=beta_var,
            epsilon=epsilon_var,
            name=out_name,
            before_op=add_op2,
        )

    add_op2.enclosing_block.replace_uses_of_var_after_op(
        anchor_op=add_op2, old_var=add_op2.outputs[0], new_var=x
    )
    # Remove all the ops at once
    block.remove_ops(all_ops)
    return True


def fuse_layernorm_or_instancenorm_block(block):
    fusion_status = False
    for i, op in enumerate(list(block.operations)):
        for b in op.blocks:
            block_changed = True
            while block_changed:
                block_changed = fuse_layernorm_or_instancenorm_block(b)
        if len(op.blocks) > 0:
            # This op can't be pow
            continue

        # start pattern match if reduce_mean op is encountered
        if op.op_type == "reduce_mean":
            with block:
                fusion_status = try_to_transform(op, block)
            # has to break as the downstream iterator is affected.
            if fusion_status:
                return fusion_status
    return fusion_status


@register_pass(namespace="common")
def fuse_layernorm_or_instancenorm(prog):
    """
    Identify the pattern:

    y = gamma * (x - mean) / sqrt(variance + epsilon) + beta

    y = x * [gamma * rsqrt(variance + eps)] + (beta - mean * [gamma * rsqrt(variance + eps)])

    [....] ---> reduce_mean --->  sub -->square -->reduce_mean --> add(epsilon)-->rsqrt
       |             |             ^                                                |
       |             |             |                                                |
       |             |             |                                                V
       |---------------------------                                               mul (gamma)
       |             |                                                              |
       |             |                                                      --------|---------
       |             |                                                      |                |
       |             |                                                      |                V
       |             |------------------------------------------------------------------->  mul
       |                                                                    |                |
       |                                                                    |                |
       |                                                                    V                |
       |-----------------------------------------------------------------> mul               |
                                                                            |                |
                                                                            |                V
                                                                            |              sub (beta) --> add --> [...]
                                                                            |                              ^
                                                                            |                              |
                                                                            |-------------------------------


    This pattern corresponds to either layer_norm or instance_norm.

    It is instance_norm if all of the following are true:
        - input is rank 4
        - axes of reduce_mean is [-2, -1]
        - gamma and beta are rank 1, after squeeze

    It is layer_norm if all of the following are true:
        - axes is either [-1] or [-1,-2] or [-1,-2,-3] and so on
        - rank of gamma and beta is equal to the length of the axes
    """
    for f_name, f in prog.functions.items():
        block_changed = True
        while block_changed:
            block_changed = fuse_layernorm_or_instancenorm_block(f)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _
from ..basic_graph_ops import delete_node
import logging
import sys

sys.setrecursionlimit(5000)  # increase recursion limit to support convert large models


def _all_assert_leaves(gdict, nodename, memo):
    """
    Does the given node lead to only assertions?

    Args:
        gdict (dict): The node's graph.
        nodename (str): The name of the node to test.
        memo (dict): Storage for memoization.
    """
    work = [nodename]
    while True:
        assert len(work) <= len(gdict)  # If true, this algorithm is broken
        node = gdict[work.pop()]

        # Entries in memo have one of the following values for a given node:
        #  None: the node is in the stack; this node is downstream.
        #  True: the node is an assertion or leads only to assertions.
        # False: the node does not lead only to assertions.
        if not isinstance(memo.get(node.name), bool):
            memo[node.name] = None
            outputs = node.outputs
            if len(outputs) == 0:
                # Leaf node: stack shrinks
                memo[node.name] = node.op in ("Assert", "CheckNumerics")
            else:
                outputs_to_process = [n for n in outputs if n not in memo]
                if len(outputs_to_process) == 0:
                    # Non-leaf node with fully processed outputs: stack shrinks
                    memo[node.name] = all(memo[n] for n in outputs)
                else:
                    # Non-leaf node with unprocess outputs: stack grows
                    work.append(node.name)
                    work.extend(outputs_to_process)
        if len(work) == 0:
            return memo[node.name]


def delete_asserts(tfssa):
    """
    Delete all nodes that lead only to assertions.
    """
    delete_count = 0
    for f in tfssa.functions.values():
        memo = {}
        for n in f.graph:
            _all_assert_leaves(f.graph, n, memo)
        for m in memo:
            if memo[m]:
                delete_count += 1
                delete_node(f.graph, m)
    logging.debug("%d assert nodes deleted", delete_count)
    return delete_count



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import absolute_import as _
from __future__ import division as _
from __future__ import print_function as _

import logging as _logging
import os.path as _os_path

import tensorflow as _tf
from coremltools.converters.mil.frontend.tensorflow.basic_graph_ops import fill_outputs
from coremltools.converters.mil.frontend.tensorflow.load import TFLoader
from coremltools.converters.mil.frontend.tensorflow.parsed_tf_node import ParsedTFNode
from coremltools.converters.mil.frontend.tensorflow.tf_graph_pass import (
    constant_propagation,
    remove_variable_nodes,
    tensor_array_resource_removal,
    insert_get_tuple,
    delete_disconnected_nodes,
    fuse_dilation_conv,
)
from coremltools.converters.mil.frontend.tensorflow.tfssa import (
    NetworkEnsemble,
    SSAFunction,
)
from coremltools.converters.mil.frontend.tensorflow2.tf_graph_pass import (
    flatten_sub_graph_namespaces,
    rewrite_control_flow_functions,
)
from tensorflow.lite.python.util import get_grappler_config as _get_grappler_config
from tensorflow.lite.python.util import (
    run_graph_optimizations as _run_graph_optimizations,
)
from tensorflow.python.framework import dtypes as _dtypes
from tensorflow.python.framework.convert_to_constants import (
    convert_variables_to_constants_v2 as _convert_variables_to_constants_v2,
)
from tensorflow.python.framework.function_def_to_graph import (
    function_def_to_graph as _function_def_to_graph,
)
from tensorflow.python.keras.saving import saving_utils as _saving_utils
from tqdm import tqdm as _tqdm

from .converter import TF2Converter


class TF2Loader(TFLoader):
    def __init__(self, model, debug=False, **kwargs):
        """
        TensorFlow 2.x model loader.

        Parameters
        ----------
        model: Model created with TensorFlow 2.x
            One of the following model format:
                - TensorFlow tf.keras.Model object or HDF5 (.h5) file path
                - TensorFlow SavedModel directory path
                - TensorFlow list of concrete functions(s)
        debug: bool, optional. Defaults to False.
            This flag should generally be False except for debugging purposes
            for diagnosing conversion errors. Setting this flag to True will
            cause graph pass errors to be ignored, forcefully returning a
            NetworkEnsemble object.
        kwargs: dict(str, Any), optional
            Dictionary of additional arguments.
        """
        TFLoader.__init__(self, model, debug, **kwargs)

    def _graph_def_from_model(self, outputs=None):
        """Overwrites TFLoader._graph_def_from_model()"""
        msg = (
            "Expected model format: [SavedModel | [concrete_function] | "
            "tf.keras.Model | .h5], got {}"
        )
        if (
            isinstance(self.model, list)
            or isinstance(self.model, _tf.keras.Model)
            or isinstance(self.model, str)
        ):
            cfs = []
            if isinstance(self.model, list):
                cfs = self.model
            if isinstance(self.model, _tf.keras.Model):
                cfs = self._concrete_fn_from_tf_keras_or_h5(self.model)
            elif isinstance(self.model, str):
                if not _os_path.exists(self.model):
                    raise ValueError(
                        'Input model "{}" does not exist'.format(self.model)
                    )
                elif _os_path.isfile(self.model) and self.model.endswith(".h5"):
                    cfs = self._concrete_fn_from_tf_keras_or_h5(self.model)
                elif _os_path.isdir(self.model):
                    saved_model = _tf.saved_model.load(self.model)
                    sv = saved_model.signatures.values()
                    cfs = sv if isinstance(sv, list) else list(sv)
                else:
                    raise NotImplementedError(msg.format(self.model))

            graph_def = self._graph_def_from_concrete_fn(cfs)
            return self.extract_sub_graph(graph_def, outputs)
        else:
            raise NotImplementedError(msg.format(self.model))

    def _tf_ssa_from_graph_def(self, fn_name="main"):
        """Overwrites TFLoader._tf_ssa_from_graph_def()"""
        with _tf.Graph().as_default() as tf_graph:
            _tf.graph_util.import_graph_def(self._graph_def, name="")

        # sub-graphs' input shapes are required for extracting sub-graphs
        sg_input_shapes = self._populate_sub_graph_input_shapes(
            tf_graph, tf_graph._functions
        )

        # get graph_dict and sub-graphs' inputs / outputs
        graph_dict, inputs, outputs, ret = self._dict_from_graph_def(
            tf_graph, fn_name, sg_input_shapes
        )

        tf_ssa = NetworkEnsemble()
        for name, graph in graph_dict.items():
            tensor_array_resource_removal(graph)
            graph = insert_get_tuple(graph)
            graph = fill_outputs(graph)
            if name == "main":  # skip for sub-graphs as input can be also output
                delete_disconnected_nodes(graph)
            tf_ssa.functions[name] = SSAFunction(
                graph, inputs=inputs[name], outputs=outputs[name], ret=ret[name]
            )

        return tf_ssa

    def _program_from_tf_ssa(self):
        # Notes:
        # - "flatten_while_loop_namespaces" should be after "constant_propagation"
        #   as it changes node names which constant propagation pass is relying on
        #   to perform session.run(), renamed nodes are not understandable for TF.
        tf_passes = [
            constant_propagation,
            rewrite_control_flow_functions,
            flatten_sub_graph_namespaces,
            remove_variable_nodes,
            fuse_dilation_conv,
        ]

        if self.debug:
            for tf_pass in _tqdm(
                tf_passes, desc="Running TensorFlow Graph Passes", unit=" passes"
            ):
                try:
                    tf_pass(self._tf_ssa)
                except Exception as e:
                    _logging.exception('Exception in pass "{}": {}'.format(tf_pass, e))
                    _logging.info("Ignoring exception and continuing to next pass")

        else:
            for tf_pass in _tqdm(
                tf_passes, desc="Running TensorFlow Graph Passes", unit=" passes"
            ):
                tf_pass(self._tf_ssa)

        if self.debug:
            import graphviz

            dot_string = self._tf_ssa.get_dot_string(
                annotation=True, name_and_op_style=True, highlight_debug_nodes=[]
            )
            graphviz.Source(dot_string).view(
                filename="/tmp/ssa_after_tf_passes", cleanup=True
            )

        converter = TF2Converter(self._tf_ssa, **self.kwargs)
        return converter.convert()

    def _populate_sub_graph_input_shapes(self, graph, graph_fns):
        """
        Populate function (sub-graph) input shapes from control flow op's inputs
        Note that the functions (sub-graphs) are not nested but the control flow
        ops are nested. The input shapes are used to extract sub-graphs from the
        parent graph (as the input of function_def_to_graph).

        Parameter
        ---------
        graph: tf.Graph
            TensorFlow graph.
        graph_fns: list of graph functions.
            List of TensorFlow graph functions.

        Returns
        -------
        sg_input_shapes: dict(str: list)
            Dictionary of function (sub-graph) name and input shape pairs.
        """
        sg_input_shapes = {}
        sub_graphs = []
        for op in graph.get_operations():
            if op.type not in {"StatelessIf", "If", "StatelessWhile", "While"}:
                continue

            sg1, sg2 = None, None
            if op.type in {"StatelessIf", "If"}:
                sg1 = op.get_attr("then_branch").name
                sg2 = op.get_attr("else_branch").name
            if op.type in {"StatelessWhile", "While"}:
                sg1 = op.get_attr("cond").name
                sg2 = op.get_attr("body").name

            # memorize input shapes for sub-graph conversions
            op_input_shapes = [i.get_shape() for i in op.inputs]
            sg_input_shapes.update({sg1: op_input_shapes, sg2: op_input_shapes})
            sub_graphs += [sg1, sg2]

        for name in sub_graphs:
            sg = graph_fns.get(name)
            fn_def = sg.definition
            op_input_shapes = sg_input_shapes[name]
            op_input_shapes = op_input_shapes[-len(fn_def.signature.input_arg) :]
            fn_graph = _function_def_to_graph(fn_def, input_shapes=op_input_shapes)
            sg_input_shapes.update(
                self._populate_sub_graph_input_shapes(fn_graph, graph_fns)
            )

        return sg_input_shapes

    @staticmethod
    def _dict_from_graph_def(graph, fn_name="main", sg_input_shapes=None):
        """
        Loads a tf.Graph and transform it into dictionary of ParsedTFNodes.
        Potentially contains multiple functions, in such case, recursively
        resolve functions (sub-graphs).

        Parameters
        ----------
        graph: tf.Graph
            TensorFlow graph.
        fn_name: str, optional, defaults to 'main'
            Function name of the graph.
        sg_input_shapes: dict(str: list)
            Dictionary of name and input shapes for functions / sub-graphs.

        Returns
        -------
        dict(str: dict(str: ParsedTFNode))
            Dictionary of function name and dictionary of node name and
            ParsedTFNode object.
        """
        graph_dict = {fn_name: {}}
        graph_inputs = {fn_name: []}
        graph_outputs = {fn_name: []}
        graph_ret = {fn_name: {}}

        for op in graph.get_operations():
            graph_dict[fn_name].update({op.name: ParsedTFNode(op.node_def)})

        for name, sg in graph._functions.items():
            sg_def = sg.definition
            if name in sg_input_shapes:
                input_shapes = sg_input_shapes[name]
                input_shapes = input_shapes[-len(sg_def.signature.input_arg):]
                fn_graph = _function_def_to_graph(sg_def, input_shapes=input_shapes)

                graph_dict.update(
                    TF2Loader._dict_from_graph_def(fn_graph, name, sg_input_shapes)[0]
                )
                graph_inputs.update({name: [t.name.split(":")[0] for t in fn_graph.inputs]})
                graph_outputs.update(
                    {name: [t.name.split(":")[0] for t in fn_graph.outputs]}
                )

                # ret is a mapping from the output arg names from `signature` to the
                # outputs from `node_def` that should be returned by the function.
                graph_ret.update({name: sg_def.ret})

        return graph_dict, graph_inputs, graph_outputs, graph_ret

    @staticmethod
    def _concrete_fn_from_tf_keras_or_h5(keras_model):
        if isinstance(keras_model, _tf.keras.Model):
            input_signature = _saving_utils.model_input_signature(
                keras_model, keep_original_batch_size=True
            )
            fn = _saving_utils.trace_model_call(keras_model, input_signature)
        else:
            keras_model = _tf.keras.models.load_model(keras_model)
            input_signature = _saving_utils.model_input_signature(
                keras_model, keep_original_batch_size=True
            )
            fn = _saving_utils.trace_model_call(keras_model, input_signature)
        return [fn.get_concrete_function()]

    @staticmethod
    def _graph_def_from_concrete_fn(cfs):
        if len(cfs) != 1:
            raise NotImplementedError("Only a single concrete function is supported.")

        frozen_fn = _convert_variables_to_constants_v2(cfs[0], lower_control_flow=False)
        graph_def = frozen_fn.graph.as_graph_def(add_shapes=True)

        # run a Grappler's constant folding pass.
        fn_inputs = [t for t in frozen_fn.inputs if t.dtype != _dtypes.resource]
        graph_def = _run_graph_optimizations(
            graph_def,
            fn_inputs,
            frozen_fn.outputs,
            config=_get_grappler_config(["constfold", "dependency", "debug_stripper"]),
            graph=frozen_fn.graph,
        )
        return graph_def



from __future__ import absolute_import as _
from __future__ import division as _
from __future__ import print_function as _
from __future__ import unicode_literals as _

from coremltools._deps import _HAS_ONNX, MSG_ONNX_NOT_FOUND, _IS_MACOS
import unittest
import numpy as np
import numpy.testing as npt  # type: ignore
import numpy.random as npr

from PIL import Image  # type: ignore

if _HAS_ONNX:
    import onnx
    from coremltools.converters.onnx import convert
    from ._test_utils import _onnx_create_single_node_model


@unittest.skipUnless(_HAS_ONNX, MSG_ONNX_NOT_FOUND)
class ConvertTest(unittest.TestCase):
    def setUp(self):  # type: () -> None
        self.img_arr = np.uint8(npr.rand(224, 224, 3) * 255)  # type: ignore
        self.img = Image.fromarray(np.uint8(self.img_arr))  # type: ignore
        self.img_arr = np.float32(self.img_arr)  # type: ignore
        self.onnx_model = _onnx_create_single_node_model(
            "Relu", [(3, 224, 224)], [(3, 224, 224)]
        )
        self.input_names = [i.name for i in self.onnx_model.graph.input]
        self.output_names = [o.name for o in self.onnx_model.graph.output]

    def test_convert_image_input(self):  # type: () -> None
        coreml_model = convert(self.onnx_model, image_input_names=self.input_names)
        spec = coreml_model.get_spec()
        for input_ in spec.description.input:
            self.assertEqual(input_.type.WhichOneof("Type"), "imageType")

    def test_convert_image_output(self):  # type: () -> None
        coreml_model = convert(self.onnx_model, image_output_names=self.output_names)
        spec = coreml_model.get_spec()
        for output in spec.description.output:
            self.assertEqual(output.type.WhichOneof("Type"), "imageType")

    def test_convert_image_input_preprocess(self):  # type: () -> None
        bias = np.array([100, 90, 80])
        coreml_model = convert(
            self.onnx_model,
            image_input_names=self.input_names,
            preprocessing_args={
                "is_bgr": True,
                "blue_bias": bias[0],
                "green_bias": bias[1],
                "red_bias": bias[2],
            },
        )

        if _IS_MACOS:
            output = coreml_model.predict({self.input_names[0]: self.img})[
                self.output_names[0]
            ]

            expected_output = self.img_arr[:, :, ::-1].transpose((2, 0, 1))
            expected_output[0] = expected_output[0] + bias[0]
            expected_output[1] = expected_output[1] + bias[1]
            expected_output[2] = expected_output[2] + bias[2]
            npt.assert_equal(output.flatten(), expected_output.flatten())

    def test_convert_image_output_bgr(self):  # type: () -> None
        coreml_model = convert(
            self.onnx_model,
            image_input_names=self.input_names,
            image_output_names=self.output_names,
            deprocessing_args={"is_bgr": True},
        )

        if _IS_MACOS:
            output = coreml_model.predict({self.input_names[0]: self.img})[
                self.output_names[0]
            ]
            output = np.array(output)[:, :, :3].transpose((2, 0, 1))
            expected_output = self.img_arr[:, :, ::-1].transpose((2, 0, 1))
            npt.assert_equal(output, expected_output)
